from __future__ import annotations
from boto3 import Session as BotoSession
from collections import namedtuple
import configparser
from datetime import datetime, timedelta
import io
import os
import re
import subprocess
import sys
from typing import List, Optional, Tuple
from hms_utils.terminal_utils import terminal_color
from hms_utils.threading_utils import run_concurrently
from hms_utils.version_utils import get_version

# ----------------------------------------------------------------------------------------------------------------------
# Convenience utility to view/manage SSO/Okta-based AWS credentials, defined in the ~/.aws/config file.
# Mostly just viewing features; the only "manage" feature being setting the default profile;
# and also refreshing credentials (i.e. logging in via aws sso login). Usage:
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# usage: python aws_env.py [profile-name-pattern] [nocheck]
#        python aws_env.py default [profile-name] | aws_env nodefault
#        python aws_env.py current [profile-name] | aws_env nocurrent
#        python aws_env.py refresh [profile-name]
# ----------------------------------------------------------------------------------------------------------------------

AWS_CONFIG_FILE_PATH = os.path.expanduser("~/.aws/config")
AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME = "AWS_PROFILE"
AWS_COMMAND_PATH = "/usr/local/bin/aws"
AWS_DEFAULT_SECTION_NAME = "default"


class AwsProfiles(List[object]):

    def __init__(self, _default: Optional[object] = None, _current: Optional[object] = None) -> None:
        self.default = _default
        self.current = _current
        pass

    def filter(self, aws_profile_name_pattern: str) -> AwsProfiles:
        aws_profiles_filtered = AwsProfiles(self.default)
        if isinstance(aws_profile_name_pattern, str) and aws_profile_name_pattern:
            aws_profile_name_pattern = aws_profile_name_pattern.lower()
            for aws_profile in self:
                if aws_profile_name_pattern in aws_profile.name.lower():
                    aws_profiles_filtered.append(aws_profile)
        return aws_profiles_filtered

    def find(self, aws_profile_name: str) -> Optional[object]:
        if isinstance(aws_profile_name, str) and aws_profile_name:
            aws_profile_name = aws_profile_name.lower()
            for aws_profile in self:
                if aws_profile.name.lower() == aws_profile_name:
                    return aws_profile
        return None

    @staticmethod
    def read() -> AwsProfiles:

        global AWS_CONFIG_FILE_PATH, AWS_DEFAULT_SECTION_NAME

        def create_aws_profile(name: str, account_number: str, default: bool = False, current: bool = False):
            return namedtuple("aws_profile", ["name", "account", "default", "current"])(
                    name, account_number, default is True, name == get_current_aws_profile_name())

        aws_config = configparser.ConfigParser()
        aws_config.read(AWS_CONFIG_FILE_PATH)
        aws_profiles = AwsProfiles()
        aws_default_profile = None

        if aws_default_section_name := aws_config_section_name(aws_config, AWS_DEFAULT_SECTION_NAME):
            if aws_default_profile_account_number := aws_config.get(aws_default_section_name, "sso_account_id"):
                aws_default_profile_name = (aws_config.get(aws_default_section_name, "sso_session") or
                                            AWS_DEFAULT_SECTION_NAME)
                aws_default_profile = create_aws_profile(aws_default_profile_name,
                                                         aws_default_profile_account_number, default=True)
                aws_profiles.default = aws_default_profile

        ndefaults = 0
        for aws_section_name in aws_config.sections():
            aws_section_name = normalize_string(aws_section_name)
            if not aws_section_name.startswith("profile "):
                continue
            aws_profile_name = aws_section_name[len("profile "):].strip()
            aws_account_number = aws_config.get(aws_section_name, "sso_account_id", fallback=None)
            aws_default = (aws_account_number == aws_default_profile.account) if aws_default_profile else False
            if aws_default:
                ndefaults += 1
            aws_profile = create_aws_profile(aws_profile_name, aws_account_number, default=aws_default)
            aws_profiles.append(aws_profile)
            if aws_profile.current:
                aws_profiles.current = aws_profile

        if ndefaults > 1:
            print(f"WARNING: AWS default section refers to more than one account number: {aws_default_profile.account}")
        elif (ndefaults == 0) and aws_default_profile:
            print(f"WARNING: AWS default section references non-existent account number: {aws_default_profile.account}")

        return aws_profiles._sorted()

    def _sorted(self) -> AwsProfiles:
        aws_profiles_sorted = AwsProfiles(_default=self.default, _current=self.current)
        for aws_profile in sorted(self, key=lambda item: item.name):
            aws_profiles_sorted.append(aws_profile)
        return aws_profiles_sorted


class CHAR:
    check = "✓"
    xmark = "✗"
    rarrow = "▶"
    rarrow_hollow = "▷"
    larrow = "◀"
    bullet = "•"
    bullet_hollow = "◦"


def verify_aws_account(aws_profile: object) -> Tuple[bool, Optional[str], Optional[str], Optional[str]]:
    aws_profile_name = aws_profile.name
    try:
        boto_session = BotoSession(profile_name=aws_profile_name)
        access_key_id = None
        expiration_time = None
        valid_duration = None
        try:
            credentials = boto_session.get_credentials()
            access_key_id = credentials.access_key
            try:
                # The credentials._expiry_time moves forward in time
                # as it is used; this is expected (I'm pretty sure).
                expiration_time = credentials._expiry_time.astimezone()
                valid_duration = get_duration(expiration_time - datetime.now(tz=expiration_time.tzinfo))
                expiration_time = expiration_time.strftime("%Y-%m-%d %H:%M:%S")
            except Exception:
                pass
        except Exception:
            pass
        boto_sts = boto_session.client("sts")
        account_number = boto_sts.get_caller_identity()["Account"]
        if isinstance(account_number, str) and (len(account_number) > 0):
            return True, access_key_id, expiration_time, valid_duration
    except Exception:
        pass
    return False, None, None, None


def get_current_aws_profile_name() -> Optional[str]:
    global AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME
    return os.environ.get(AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME, None)


def perform_login(aws_profile_name: str, verbose: bool = False) -> None:
    global AWS_COMMAND_PATH
    command = [AWS_COMMAND_PATH, "sso", "login", "--profile", aws_profile_name]
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    try:
        look_for_code = False
        while True:
            if (not (output := process.stdout.readline())) and (process.poll() is not None):
                break
            if (look_for_code is not None) and (output := output.strip()):
                if "enter the code" in output.lower():
                    look_for_code = True
                elif look_for_code:
                    print(f"Confirm this code in your browser to login to {aws_profile_name}: {output.strip()} ... ",
                          end="", flush=True)
                    look_for_code = None
        if stderr_output := process.stderr.read().strip():
            print(f"\nNot confirmed or login failed: {aws_profile_name}")
            if verbose is True:
                print(f"\nError output on login ({aws_profile_name}):")
                print(stderr_output)
        else:
            print("OK")
        process.wait()
    except Exception as e:
        print(f"\nNot confirmed or login failed: {aws_profile_name}")
        if verbose is True:
            print(f"EXCEPTION: {str(e)}")


def print_aws_profile_line(aws_profile: object, verified_result: Optional[tuple] = None, nocheck: bool = False,
                           current: Optional[str] = None, login: bool = False, verbose: bool = False) -> bool:
    global AWS_DEFAULT_SECTION_NAME
    verified = False if not verified_result else verified_result[0]
    if login:
        perform_login(aws_profile.name, verbose=verbose)
    aws_profile_name_current = get_current_aws_profile_name()
    if current and not aws_profile.current:
        return False
    if ((aws_profile_name_current == aws_profile.name) or aws_profile.default or
        (aws_profile.name == AWS_DEFAULT_SECTION_NAME)):  # noqa
        line = f"{CHAR.rarrow} {aws_profile.name}: "
    else:
        line = f"{CHAR.rarrow_hollow} {aws_profile.name}: "
    if aws_profile.account:
        line += f"{aws_profile.account}"
    else:
        line += "<no account number>"
    if not nocheck:
        if verified_result:
            verified, access_key_id, expiration_time, valid_duration = verified_result
        else:
            verified, access_key_id, expiration_time, valid_duration = verify_aws_account(aws_profile)
        if verified:
            line += f" {CHAR.check}"
            if access_key_id and verbose:
                line += f" {access_key_id}"
            if expiration_time and verbose:
                if verbose:
                    line += f" {CHAR.rarrow_hollow} {expiration_time}"
                else:
                    line += f" {expiration_time}"
            if valid_duration and verbose:
                line += f" {CHAR.bullet_hollow} {valid_duration}"
        else:
            line += f" {CHAR.xmark}"
    if aws_profile.default:
        line += f" {CHAR.larrow} default"
    if aws_profile_name_current == aws_profile.name:
        AwsProfiles().read().default
        if ((not aws_profile.default) and
            (aws_default_profile := AwsProfiles().read().default) and
            (aws_default_profile.name != aws_profile.name)):  # noqa
            # If both current and default set and current is not the same as default then
            # make the current one stand out, as this is the one that will be used over default.
            line += f" {CHAR.larrow} {terminal_color('current', 'red', bold=True, underline=True)}"
        else:
            line += f" {CHAR.larrow} current"
    print(line)
    return verified


def set_default_profile(aws_profile: object, auto_confirm: bool = False) -> bool:
    global AWS_CONFIG_FILE_PATH, AWS_DEFAULT_SECTION_NAME
    aws_profile_name = aws_profile.name
    aws_config = configparser.ConfigParser()
    aws_config.read(AWS_CONFIG_FILE_PATH)
    if aws_config_section_name(aws_config, AWS_DEFAULT_SECTION_NAME):
        if aws_config.get(AWS_DEFAULT_SECTION_NAME, "sso_account_id", fallback=None) == aws_profile.account:
            print(f"AWS profile is already the default: {aws_profile.name}")
            return False
    if aws_profile_section_name := aws_config_section_name(aws_config, f"profile {aws_profile_name}"):
        if aws_default_section_name := aws_config_section_name(aws_config, AWS_DEFAULT_SECTION_NAME):
            aws_config.remove_section(aws_default_section_name)
        aws_config.add_section(AWS_DEFAULT_SECTION_NAME)
        for name, value in aws_config.items(aws_profile_section_name):
            aws_config.set(AWS_DEFAULT_SECTION_NAME, name, value)
        if aws_session_section_name := aws_config_section_name(aws_config, f"sso-session {aws_profile_name}"):
            if (aws_default_session_section_name :=
                aws_config_section_name(aws_config, f"sso-session {AWS_DEFAULT_SECTION_NAME}")):  # noqa
                aws_config.remove_section(aws_default_session_section_name)
            aws_config.add_section(f"sso-session {AWS_DEFAULT_SECTION_NAME}")
            for name, value in aws_config.items(aws_session_section_name):
                aws_config.set(f"sso-session {AWS_DEFAULT_SECTION_NAME}", name, value)
    if auto_confirm is not True:
        if not confirm(f"Update AWS config file with new default ({aws_profile_name}): {AWS_CONFIG_FILE_PATH} "):
            return False
    with open(AWS_CONFIG_FILE_PATH, "w") as f:
        aws_config.write(f)
    return True


def remove_default_profile(default: Optional[object] = None, auto_confirm: bool = False) -> None:
    global AWS_CONFIG_FILE_PATH
    aws_config = configparser.ConfigParser()
    aws_config.read(AWS_CONFIG_FILE_PATH)
    if not (aws_default_section_name := aws_config_section_name(aws_config, AWS_DEFAULT_SECTION_NAME)):
        print("No default AWS profile to remove.")
        return
    if aws_default_section_name := aws_config_section_name(aws_config, AWS_DEFAULT_SECTION_NAME):
        if auto_confirm is not True:
            if not confirm(f"Update AWS config file to remove default"
                           f" profile ({default.name}): {AWS_CONFIG_FILE_PATH} "):
                return
        aws_config.remove_section(aws_default_section_name)
        if aws_session_section_name := aws_config_section_name(aws_config, f"sso-session {AWS_DEFAULT_SECTION_NAME}"):
            aws_config.remove_section(aws_session_section_name)
        with open(AWS_CONFIG_FILE_PATH, "w") as f:
            aws_config.write(f)


def aws_config_section_name(aws_config: configparser.ConfigParser, aws_section_name: str) -> Optional[str]:
    # Use this rather than aws_config.has_section to handle weird spacing in ~/.aws/config;
    # returns the actual section name as it is in ~/.aws/config or None if not found.
    if (isinstance(aws_config, configparser.ConfigParser) and
        isinstance(aws_section_name := normalize_string(aws_section_name).lower(), str)):  # noqa
        for aws_config_section_name in aws_config.sections():
            if normalize_string(aws_config_section_name).lower() == aws_section_name:
                return aws_config_section_name
    return None


def get_duration(value: timedelta) -> str:
    total_seconds = int(value.total_seconds())
    hours, hours_remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(hours_remainder, 60)
    return f"{hours}:{minutes:02}:{seconds:02}"


def normalize_string(value: str) -> str:
    return re.sub(r"\s+", " ", value).strip() if isinstance(value, str) else ""


def confirm(message: str) -> bool:
    return input(f"{message}? ").lower() in ["y", "yes"]


def usage(status: int = 1) -> None:
    print("usage: python aws_env.py [profile-name-pattern] [nocheck]")
    print("       python aws_env.py login [profile-name]")
    print("       python aws_env.py default [profile-name]")
    print("       python aws_env.py nodefault")
    print("       python aws_env.py current [profile-name]")
    print("       python aws_env.py nocurrent")
    sys.exit(status)


def main() -> None:

    global AWS_CONFIG_FILE_PATH, AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME

    nocheck = False
    default = False
    nodefault = False
    current = False
    nocurrent = False
    noasync = False
    login = False
    yes = False
    profile_name_pattern = None
    verbose = False
    current_export_file = None
    post_current_export_file = False

    if os.environ.get("AWS_CONFIG"):
        AWS_CONFIG_FILE_PATH = os.environ.get("AWS_CONFIG")

    argi = 0
    args = []
    while argi < len(argv := sys.argv[1:]):
        arg = argv[argi]
        if (arg == "--config") or (arg == "-config") or (arg == "--conf") or (arg == "-conf"):
            argi += 1
            if argi >= len(argv):
                usage()
            AWS_CONFIG_FILE_PATH = argv[argi]
        else:
            args.append(arg)
        argi += 1

    if not os.path.exists(AWS_CONFIG_FILE_PATH):
        print(f"AWS config file not found: {AWS_CONFIG_FILE_PATH}")
        sys.exit(2)
    if not (aws_profiles := AwsProfiles.read()):
        print(f"AWS config file contains NO profiles: {AWS_CONFIG_FILE_PATH}")
        sys.exit(3)

    # Since this is a convenience script, for convenience allow
    # a multitude of option variations so we don't need to remember.
    argi = 0
    while argi < len(args):
        arg = args[argi]
        if (arg == "--shell") or (arg == "-shell"):
            print(os.path.join(os.path.dirname(os.path.abspath(__file__)), "aws_env.sh"))
            exit(1)
        elif ((arg == "--nocheck") or (arg == "-nocheck") or (arg == "nocheck") or
              (arg == "--nc") or (arg == "-nc") or (arg == "nc") or (arg == "--n") or (arg == "-n") or
              (arg == "--quick") or (arg == "-quick") or (arg == "quick") or (arg == "--q") or (arg == "-q")):  # noqa
            nocheck = True
        elif ((arg == "--active") or (arg == "-active") or (arg == "active") or
              (arg == "--activate") or (arg == "-activate") or (arg == "activate") or
              (arg == "--current") or (arg == "-current") or (arg == "current") or (arg == "--c") or (arg == "-c")):
            current = True
        elif ((arg == "--noactive") or (arg == "-noactive") or (arg == "active") or
              (arg == "--deactive") or (arg == "-deactive") or (arg == "deactive") or
              (arg == "--deactivate") or (arg == "-deactivate") or (arg == "deactivate") or
              (arg == "--nocurrent") or (arg == "-nocurrent") or (arg == "nocurrent")):
            nocurrent = True
        elif ((arg == "--refresh") or (arg == "-refresh") or (arg == "refresh") or (arg == "--r") or (arg == "-r") or
              (arg == "--login") or (arg == "-login") or (arg == "login")):
            login = True
        elif (arg == "--default") or (arg == "-default") or (arg == "default") or (arg == "--d") or (arg == "-d"):
            default = True
        elif (arg == "--nodefault") or (arg == "-nodefault") or (arg == "nodefault"):
            nodefault = True
        elif arg in ["--noasync", "-noasync", "noasync"]:
            noasync = True
        elif (arg == "--yes") or (arg == "-yes") or (arg == "yes") or (arg == "--y") or (arg == "-y"):
            yes = True
        elif (arg == "--verbose") or (arg == "-verbose") or (arg == "verbose") or (arg == "--v") or (arg == "-v"):
            verbose = True
        elif (arg == "--help") or (arg == "-help") or (arg == "help") or (arg == "--h") or (arg == "-h"):
            usage(0)
        elif (arg == "--current-export-file"):  # to support aws_env.sh script
            if (argi := argi + 1) >= len(args):
                usage(1)
            current_export_file = args[argi]
        elif (arg == "--post-current-export-file"):  # to support aws_env.sh script
            post_current_export_file = args[argi]
        elif arg in ["--version", "-version"]:
            print(f"hms-utils version: {get_version()}")
            exit(0)
        elif arg.startswith("-"):
            usage(1)
        elif profile_name_pattern:
            usage(1)
        else:
            profile_name_pattern = arg
        argi += 1

    if default:
        if nodefault or current or nocurrent:
            usage()
    if nodefault:
        if default or current or nocurrent or login:
            usage()
    if current:
        if default or nodefault or nocurrent:
            usage()
    if nocurrent:
        if current or default or nodefault or login:
            usage()

    if profile_name_pattern:
        if nodefault:
            print("With the --nodefault option (to remove the default AWS profile) do not specify an AWS profile name.")
            sys.exit(5)
        elif nocurrent:
            if not aws_profiles.current:
                print("With the --nocurrent option (to deactivate the"
                      f" current AWS profile) do not specify an AWS profile name.")
            sys.exit(5)
        elif not (aws_profiles_selected := aws_profiles.filter(profile_name_pattern)):
            print(f"AWS profile name ({profile_name_pattern})"
                  f" does not match any profiles in AWS config: {AWS_CONFIG_FILE_PATH}")
            for aws_profile in aws_profiles:
                print_aws_profile_line(aws_profile, nocheck=True)
            sys.exit(6)
        elif default:
            if len(aws_profiles_selected) != 1:
                print(f"The --default option requires a SINGLE existing AWS profile name (multiple matches: "
                      f"{', '.join([item.name for item in aws_profiles_selected])}).")
                sys.exit(7)
            if aws_profiles_selected[0].name != profile_name_pattern:
                print(f"The --default option requires an EXACT matching"
                      f" AWS profile name ({aws_profiles_selected[0].name}).")
                sys.exit(8)
        elif current:
            if len(aws_profiles_selected) != 1:
                print(f"The --current option requires a SINGLE existing AWS profile name (multiple matches: "
                      f"{', '.join([item.name for item in aws_profiles_selected])}).")
                sys.exit(9)
            if aws_profiles_selected[0].name != profile_name_pattern:
                print(f"The --current option requires an EXACT matching"
                      f" AWS profile name ({aws_profiles_selected[0].name}).")
                sys.exit(10)
    elif default:
        if not aws_profiles.default:
            print("No AWS profile is currently the default.")
            print("Use the --default option to set the default profile to an existing AWS profile name.")
            sys.exit(11)
        verified = print_aws_profile_line(aws_profiles.default, nocheck=nocheck, login=login, verbose=verbose)
        sys.exit(0 if verified or nocheck else 1)
    elif current:
        if not aws_profiles.current:
            print("No AWS profile is currently active.")
            print("Use the --current option to set the active profile to an existing AWS profile name.")
            sys.exit(12)
        verified = print_aws_profile_line(aws_profiles.current, nocheck=nocheck, login=login, verbose=verbose)
        sys.exit(0 if verified or nocheck else 1)
    else:
        aws_profiles_selected = aws_profiles

    if default:
        aws_profile_default = aws_profiles_selected[0]
        if set_default_profile(aws_profile_default, auto_confirm=yes):
            aws_profile_default = AwsProfiles.read().find(aws_profile_default.name)
        verified = print_aws_profile_line(aws_profile_default,
                                          nocheck=nocheck, current=current, login=login, verbose=verbose)
        sys.exit(0 if verified or nocheck else 1)
    elif nodefault:
        remove_default_profile(default=aws_profiles.default, auto_confirm=yes)
        sys.exit(0)
    elif nocurrent:
        if not aws_profiles.current:
            if not post_current_export_file:
                print("No current AWS profile to deactivate.")
            exit(0)
        if current_export_file and not os.path.exists(current_export_file):
            with io.open(current_export_file, "w") as f:
                f.write(f"unset {AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME}")
        else:
            print(f"To deactivate the current AWS profile ({aws_profiles.current.name}) execute:"
                  f" unset {AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME}")
        sys.exit(0)
    elif current:
        if (aws_profile_current := aws_profiles_selected[0]).current:
            if not post_current_export_file:
                print(f"This AWS profile is already currently active: {aws_profile_current.name}")
            verified = print_aws_profile_line(aws_profile_current, nocheck=nocheck, login=login, verbose=verbose)
            sys.exit(0 if verified or nocheck else 1)
        else:
            if current_export_file and not os.path.exists(current_export_file):
                with io.open(current_export_file, "w") as f:
                    f.write(f"export {AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME}={aws_profile_current.name}")
            else:
                print(f"To make this AWS profile ({arg}) current execute:"
                      f" export {AWS_PROFILE_ENVIRONMENT_VARIABLE_NAME}={aws_profile_current.name}")
            sys.exit(0)

    if (len(aws_profiles_selected) > 1) or verbose:
        print("AWS profiles", end="")
        if profile_name_pattern:
            print(" matching '", end="")
            print(profile_name_pattern, end="")
            print("'", end="")
        print(f" in config: {AWS_CONFIG_FILE_PATH}")
    if (noasync is not True) and (len(aws_profiles_selected) > 1):
        aws_profile_results = {}
        def function(aws_profile: object):  # noqa
            nonlocal aws_profile_results
            aws_profile_results[aws_profile] = verify_aws_account(aws_profile)
        run_concurrently([lambda item=item: function(item) for item in aws_profiles_selected], nthreads=12)
        for aws_profile in sorted(aws_profile_results, key=lambda item: item.name):
            print_aws_profile_line(aws_profile, nocheck=nocheck, current=current, login=login, verbose=verbose,
                                   verified_result=aws_profile_results[aws_profile])
    else:
        for aws_profile in aws_profiles_selected:
            verified = print_aws_profile_line(aws_profile, nocheck=nocheck,
                                              current=current, login=login, verbose=verbose)
            if len(aws_profiles_selected) == 1:
                # If only a single profile selected then make the exit status correspond to its verified state.
                sys.exit(0 if verified or nocheck else 1)

    sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations
import sys
from typing import Any, Callable, List, Optional, Type, Union
from uuid import uuid4 as uuid
from hms_utils.chars import chars
from hms_utils.dictionary_utils import sort_dictionary
from hms_utils.type_utils import (
    to_bool, to_float, to_integer, to_flattened_list, to_non_empty_string_list, to_string_list)


class Argv:

    BOOLEAN = 0x0001
    FLOAT = 0x0002
    FLOATS = 0x0004
    INTEGER = 0x0008
    INTEGERS = 0x0010
    STRING = 0x0020
    STRINGS = 0x0040
    DEFAULT = 0x0080
    DEFAULTS = 0x0100
    OPTIONAL = 0x0200
    REQUIRED = 0x0400

    _ESCAPE_OPTION = "--"
    _FUZZY_OPTION_PREFIX = "-"
    _FUZZY_OPTION_PREFIX_LEN = len(_FUZZY_OPTION_PREFIX)
    _OPTION_PREFIX = "--"
    _OPTION_PREFIX_LEN = len(_OPTION_PREFIX)
    _DEFAULT_VALUE_DELIMITER = ":__default__:"
    _RULE_DELIMITER = "__@@@:@@@__"
    _RULE_PREFIX = f"__rule__{_RULE_DELIMITER}"
    _RULE_PREFIX_LEN = len(_RULE_PREFIX)
    _NO_VALUE = object()

    class MistypedValue(str):
        pass

    class _Arg(str):

        def __new__(cls, value: str, argv: Argv) -> None:
            null = value is None
            value = (value.strip() if argv._strip else value) if isinstance(value, str) else ""
            value = super().__new__(cls, value)
            value._argv = argv
            value._null = null
            return value

        @property
        def is_option(self) -> bool:
            return self._argv._is_option(self)

        @property
        def is_null(self) -> bool:
            return self._null

        @property
        def is_not_null(self) -> bool:
            return not self._null

        def is_any_of(self, options) -> bool:
            if self._argv._escaping:
                return False
            return self._argv._is_option_any_of(self, options)

        def set_value_boolean(self, option: Argv._Option) -> bool:
            if isinstance(option, str): option = Argv._Option(option)  # noqa
            if self.is_any_of(option._options):
                setattr(self._argv._values, option._property_name, True)
                return True
            return False

        def set_value_string(self, option: Argv._Option) -> bool:
            return self._set_value_property(option)

        def set_value_strings(self, option: Argv._Option) -> bool:
            return self._set_value_properties(option)

        def set_value_integer(self, option: Argv._Option) -> bool:
            return self._set_value_property(option, convert=to_integer)

        def set_value_integers(self, option: Argv._Option) -> bool:
            return self._set_value_properties(option, convert=to_integer)

        def set_value_float(self, option: Argv._Option) -> bool:
            return self._set_value_property(option, convert=to_float)

        def set_value_floats(self, option: Argv._Option) -> bool:
            return self._set_value_properties(option, convert=to_float)

        def set_default_value_string(self, option: Argv._Option) -> bool:
            return self._set_default_value_property(option)

        def set_default_value_strings(self, option: Argv._Option) -> bool:
            return self._set_default_value_properties(option)

        def set_default_value_integer(self, option: Argv._Option) -> bool:
            return self._set_default_value_property(option, convert=to_integer)

        def set_default_value_integers(self, option: Argv._Option) -> bool:
            return self._set_default_value_properties(option, convert=to_integer)

        def set_default_value_float(self, option: Argv._Option) -> bool:
            return self._set_default_value_property(option, convert=to_float)

        def set_default_value_floats(self, option: Argv._Option) -> bool:
            return self._set_default_value_properties(option, convert=to_float)

        def _set_value_property(self, option: Argv._Option, convert: Optional[Callable] = None) -> bool:
            if isinstance(option, str): option = Argv._Option(option)  # noqa
            if self.is_any_of(option._options) and (not hasattr(self._argv._values, option._property_name)):
                if (value := self._argv._peek).is_not_null and ((value != Argv._ESCAPE_OPTION) or
                                                                (not value.is_option)):
                    if not ((value == Argv._ESCAPE_OPTION) and self._argv._next.is_null):
                        if (not callable(convert)) or ((value := convert(value)) is not None):
                            setattr(self._argv._values, option._property_name,
                                    value if callable(convert) else str(value))
                            self._argv._next
                            return True
                        else:
                            # Here the option and associated argument was given but it is of the wrong type;
                            # we set the value property to None and will flag it later as an error.
                            setattr(self._argv._values, option._property_name, Argv.MistypedValue(self._argv._peek))
                            self._argv._next
                            return True
                # Here the option was given but no associated argument as required;
                # we set the value property to None and will flag it later as an error.
                setattr(self._argv._values, option._property_name, Argv._NO_VALUE)
                return True
            return False

        def _set_value_properties(self, option: Union[Argv._Option, str], convert: Optional[Callable] = None) -> bool:
            if isinstance(option, str): option = Argv._Option(option)  # noqa
            if not callable(convert):
                convert = lambda value: str(value)  # noqa
            if self.is_any_of(option._options):
                values = []
                if (value := getattr(self._argv._values, option._property_name, None)) is not None:
                    if isinstance(value, list):
                        values[:0] = value
                    elif (value := convert(value)) is not None:
                        values.append(value)
                setattr(self._argv._values, option._property_name, values)
                while True:
                    if (value := self._argv._peek).is_null or value.is_option:
                        break
                    if ((value == Argv._ESCAPE_OPTION) and (value := self._argv._next).is_null) or value.is_option:
                        break
                    if (value := convert(value)) is None:
                        value = Argv.MistypedValue(self._argv._peek)
                        # break
                    values.append(value)
                    self._argv._next
                return True
            return False

        def _set_default_value_property(self, option: Argv._Option, convert: Optional[Callable] = None) -> bool:
            if self and (not self.is_option):
                if (value := self) == Argv._ESCAPE_OPTION:
                    if (value := self._argv._next).is_null:
                        return False
                for option in option._options:
                    option_property_name = Argv._property_name_ize(option)
                    if (not callable(convert)) or ((value := convert(value)) is not None):
                        if not hasattr(self._argv._values, option_property_name):
                            setattr(self._argv._values, option_property_name, value)
                            return True
            return False

        def _set_default_value_properties(self, option: Argv._Option, convert: Optional[Callable] = None) -> bool:
            parsed = False ; value = self  # noqa
            for option_option in option._options:
                option_property_name = Argv._property_name_ize(option_option)
                option_values = getattr(self._argv._values, option_property_name, None)
                while True:
                    if value.is_null:
                        break
                    if ((value == Argv._ESCAPE_OPTION) and (value := self._argv._next).is_null) or value.is_option:
                        break
                    if callable(convert) and ((value := convert(value)) is None):
                        break
                    if option_values is None:
                        option_values = []
                        setattr(self._argv._values, option_property_name, option_values)
                    option_values.append(value if callable(convert) else str(value))
                    parsed = True
                    if (value := self._argv._peek).is_null or value.is_option:
                        break
                    self._argv._next
            return parsed

    class _Values:
        pass

    class _OptionDefinitions:
        def __init__(self, fuzzy: bool = True) -> None:
            self._definitions = []
            self._fuzzy = fuzzy is not False
            self._rule_exactly_one_of = []
            self._rule_at_least_one_of = []
            self._rule_at_most_one_of = []
            self._rule_depends_on = []
            self._option_type_action_map = {
                # 0: Argv._Arg.set_value_string,
                0: Argv._Arg.set_value_boolean,
                Argv.BOOLEAN: Argv._Arg.set_value_boolean,
                Argv.STRING: Argv._Arg.set_value_string,
                Argv.STRINGS: Argv._Arg.set_value_strings,
                Argv.INTEGER: Argv._Arg.set_value_integer,
                Argv.INTEGERS: Argv._Arg.set_value_integers,
                Argv.FLOAT: Argv._Arg.set_value_float,
                Argv.FLOATS: Argv._Arg.set_value_floats
            }
            self._default_option_type_action_map = {
                0: Argv._Arg.set_default_value_string,
                Argv.STRING: Argv._Arg.set_default_value_string,
                Argv.STRINGS: Argv._Arg.set_default_value_string,
                Argv.INTEGER: Argv._Arg.set_default_value_integer,
                Argv.INTEGERS: Argv._Arg.set_default_value_integer,
                Argv.FLOAT: Argv._Arg.set_default_value_float,
                Argv.FLOATS: Argv._Arg.set_default_value_float
            }
            self._defaults_option_type_action_map = {
                0: Argv._Arg.set_default_value_strings,
                Argv.STRING: Argv._Arg.set_default_value_strings,
                Argv.STRINGS: Argv._Arg.set_default_value_strings,
                Argv.INTEGER: Argv._Arg.set_default_value_integers,
                Argv.INTEGERS: Argv._Arg.set_default_value_integers,
                Argv.FLOAT: Argv._Arg.set_default_value_floats,
                Argv.FLOATS: Argv._Arg.set_default_value_floats
            }

        def define_option(self, option_type: int, options: List[str]) -> None:
            if isinstance(option_type, int) and isinstance(options, list) and options:
                option_required = (option_type & Argv.REQUIRED) == Argv.REQUIRED
                if (option_type & Argv.DEFAULT) == Argv.DEFAULT:
                    option_type_action_map = self._default_option_type_action_map
                elif (option_type & Argv.DEFAULTS) == Argv.DEFAULTS:
                    option_type_action_map = self._defaults_option_type_action_map
                else:
                    option_type_action_map = self._option_type_action_map
                option_type &= ~(Argv.REQUIRED | Argv.OPTIONAL | Argv.DEFAULT | Argv.DEFAULTS)
                if action := option_type_action_map.get(option_type):
                    if (index := options[0].find("=")) > 0:
                        pass
                        default = options[0][index + 1:]
                        if (option_type & Argv.BOOLEAN):
                            default = to_bool(default)
                        elif (option_type & Argv.INTEGER):
                            default = to_integer(default)
                        elif (option_type & Argv.FLOAT):
                            default = to_float(default)
                        options[0] = options[0][0:index]
                    elif (option_type & Argv.BOOLEAN):
                        # default = False
                        default = None
                        pass
                    else:
                        default = None
                    self._definitions.append(Argv._Option(
                        options=options, type=option_type, required=option_required,
                        default=default, action=action, fuzzy=self._fuzzy))

        def add_rule_exactly_one_of(self, options: List[str]) -> None:
            if options := list(set(to_non_empty_string_list(options))):
                self._rule_exactly_one_of.append(options)

        def add_rule_at_least_one_of(self, options: List[str]) -> None:
            if options := list(set(to_non_empty_string_list(options))):
                self._rule_at_least_one_of.append(options)

        def add_rule_at_most_one_of(self, options: List[str]) -> None:
            if options := list(set(to_non_empty_string_list(options))):
                self._rule_at_most_one_of.append(options)

        def add_rule_depends_on(self, options: List[str]) -> None:
            dependent_options = [] ; required_options = []  # noqa
            if options := to_flattened_list(options):
                for optioni in range(len(options)):
                    if isinstance(option := options[optioni], str) and (option := option.strip()):
                        if option.startswith(f"{Argv._RULE_PREFIX}depends_on"):
                            if dependent_options := options[:optioni]:
                                if (optioni + 1) < len(options):
                                    if required_options := options[optioni + 1:]:
                                        self._rule_depends_on.append({"depends": dependent_options,
                                                                      "on": required_options})
                                        return

    class _Option:
        def __init__(self,
                     options: Optional[List[str]] = None,
                     type: Optional[int] = None,
                     required: bool = False,
                     default: Any = None,
                     action: Optional[Callable] = None,
                     fuzzy: bool = True) -> None:
            self._options = to_non_empty_string_list(options)
            self._type = type if isinstance(type, int) else None
            self._required = required is True
            self._default = default
            self._fuzzy = fuzzy is not True
            self._action = action if callable(action) else lambda: None

        @property
        def _option_name(self) -> str:
            return self._options[0] if self._options else ""

        @property
        def _property_name(self) -> str:
            if self._options and (option := self._options[0]):
                if option.startswith(Argv._OPTION_PREFIX):
                    option = option[Argv._OPTION_PREFIX_LEN:]
                elif self._fuzzy and option.startswith(Argv._FUZZY_OPTION_PREFIX):
                    option = option[Argv._FUZZY_OPTION_PREFIX_LEN:]
                return Argv._property_name_ize(option)
            return ""

        def __eq__(self, other):
            return id(self) == id(other)

        def __hash__(self):
            return hash(id(self))

    def __init__(self, *args, argv: Optional[List[str]] = None,
                 parse: bool = True, exit: bool = True, fuzzy: bool = True,
                 strip: bool = True, skip: bool = True, escape: bool = True, delete: bool = False) -> None:
        self._argi = 0
        self._fuzzy = fuzzy is not False
        self._strip = strip is not False  # TODO: make strip = False by default
        self._escape = escape is not False
        self._escaping = False
        self._delete = delete is True
        self._values = Argv._Values()
        self._exit = (parse is True) and (exit is True)
        if (len(args) == 1) and isinstance(args[0], list):
            # Here, the given args are the actual command-line arguments to process/parse.
            if not (isinstance(argv, list) and argv):
                argv = args[0]
            self._option_definitions = self._process_option_definitions()
        else:
            # Here, the given args should be the definitions for processing/parsing command-line args.
            self._option_definitions = self._process_option_definitions(*args)
        # self._argv = argv if isinstance(argv, list) and argv else (sys.argv[1:] if skip is not False else sys.argv)
        if isinstance(argv, list) and argv:
            self._argv = argv
        else:
            self._argv = sys.argv[1:] if skip is not False else sys.argv
        if (parse is True) and self._option_definitions._definitions:
            self.parse()

    @property
    def list(self) -> List[str]:
        return self._argv

    @property
    def values(self) -> Argv._Values:
        return self._values

    def parse(self, *args, skip: bool = True, report: bool = True,
              exit: Optional[bool] = None, printf: Optional[Callable] = None) -> List[str]:

        if exit not in (True, False):
            exit = self._exit
        if ((len(args) == 1) and isinstance(args[0], list)) or (len(args) == 0):
            # Here, the given args are the actual command-line arguments to process/parse;
            # and this Argv object already should have the definitions for processing/parsing these.
            if not self._option_definitions:
                return None, None
            self._argv = args[0] if (len(args) > 0) else (sys.argv[1:] if (skip is not False) else sys.argv)
        else:
            # Here, the given args are the definitions for processing/parsing command-line args;
            # and this Argv object already as the actual command-line arguments to process/parse.
            self._option_definitions = self._process_option_definitions(*args)  # xyzzy

        if not self._option_definitions:
            return []

        errors = []
        missing_options = []
        missing_option_arguments = []
        mistyped_option_arguments = []
        missing_option_arguments_multiple = []
        mistyped_option_arguments_multiple = []
        duplicate_options = []
        unparsed_options = []
        rule_violations_exactly_one_of_missing = []
        rule_violations_exactly_one_of_toomany = []
        rule_violations_at_least_one_of_missing = []
        rule_violations_at_most_one_of_toomany = []
        rule_violations_depends_on = []

        # Parse each of the actual options/arguments.
        for arg in self:
            parsed = False
            for option in self._option_definitions._definitions:
                if option._action(arg, option):
                    parsed = True
                    break
            if not parsed:
                if (option := self._find_option(arg)) and hasattr(self._values, option._property_name):
                    duplicate_options.append(option._option_name)
                else:
                    unparsed_options.append(arg)

        # Flag missing arguments for argument options.
        for option in self._option_definitions._definitions:
            if (value := getattr(self._values, option._property_name, None)) is not None:
                if isinstance(value, Argv.MistypedValue):
                    mistyped_option_arguments.append((option, value))
                elif value == Argv._NO_VALUE:
                    missing_option_arguments.append(option)
                elif isinstance(value, list):
                    if not value:
                        missing_option_arguments_multiple.append(option)
                        continue
                    for item in [item for item in value if isinstance(item, Argv.MistypedValue)]:
                        mistyped_option_arguments_multiple.append((option, item))

        # Define default value properties.
        for option in self._option_definitions._definitions:
            if not hasattr(self._values, option._property_name) and (option._default is not None):
                setattr(self._values, option._property_name, option._default)

        # Enforce various auxiliary rules.
        defined_value_options = set(self._defined_value_options())

        # Enforce exactly-one-of rule.
        for rule_options in self._option_definitions._rule_exactly_one_of:
            if rule_options := set(self._find_options(rule_options)):
                if len(intersection_options := rule_options & defined_value_options) == 0:
                    # Exactly one of the specifed rule options should be specified but none are.
                    rule_violations_exactly_one_of_missing.append(
                        [option._option_name for option in rule_options])
                elif len(intersection_options) != 1:
                    # Exactly one of the specifed rule options should be specified more than one are.
                    rule_violations_exactly_one_of_toomany.append(
                        [option._option_name for option in rule_options])

        # Enforce at-least-one-of rule.
        for rule_options in self._option_definitions._rule_at_least_one_of:
            if rule_options := set(self._find_options(rule_options)):
                intersection_options = rule_options & defined_value_options
                if len(intersection_options) == 0:
                    # At least one of the specifed rule options should be specified but none are.
                    rule_violations_at_least_one_of_missing.append(
                        [option._option_name for option in rule_options])

        # Enforce at-most-one-of rule.
        for rule_options in self._option_definitions._rule_at_most_one_of:
            if rule_options := set(self._find_options(rule_options)):
                intersection_options = rule_options & defined_value_options
                if len(intersection_options) > 1:
                    # At most one of the specifed rule options should be specified but more than one are.
                    rule_violations_at_most_one_of_toomany.append(
                        [option._option_name for option in rule_options])

        # Enforce dependency rules.
        for rule in self._option_definitions._rule_depends_on:
            rule_dependent_options = to_non_empty_string_list(rule.get("depends"))
            rule_required_options = rule.get("on")
            rule_dependent_options = self._find_options(rule_dependent_options)
            rule_required_options = self._find_options(rule_required_options)
            for rule_dependent_option in rule_dependent_options:
                if hasattr(self._values, rule_dependent_option._property_name):
                    found = False
                    for rule_required_option in rule_required_options:
                        if hasattr(self._values, rule_required_option._property_name):
                            found = True
                            break
                    if not found:
                        rule_violations_depends_on.append(
                            [rule_dependent_option._option_name,
                             ", ".join([item._option_name for item in rule_required_options])])

        # Set properties to None or an appropriate default (e.g. False for boolean)
        # for any/all options/properties not specified at all; must be after above.
        for option in self._option_definitions._definitions:
            if not hasattr(self._values, option._property_name):
                if option._required and (option._option_name not in missing_options):
                    missing_options.append(option._option_name)
                value = False if option._type & ARGV.BOOLEAN else None
                setattr(self._values, option._property_name, False if option._type & ARGV.BOOLEAN else None)

        # Collect/format any errors.
        for duplicate_option in duplicate_options:
            errors.append(f"Duplicate option: {duplicate_option}")
        if unparsed_options:
            errors.append(f"Unrecognized argument"
                          f"{'s' if len(unparsed_options) > 1 else ''}: {', '.join(unparsed_options)}")
        if missing_options:
            errors.append(f"Missing required option"
                          f"{'s' if len(missing_options) > 1 else ''}: {', '.join(missing_options)}")
        for missing_option_argument in missing_option_arguments:
            errors.append(f"Missing argument for option: {missing_option_argument._option_name}")
        for missing_option_argument_multiple in missing_option_arguments_multiple:
            errors.append(f"Missing argument(s) for option: {missing_option_argument_multiple._option_name}")
        for mistyped_option, mistyped_argument in mistyped_option_arguments:
            errors.append(f"Mistyped argument for option:"
                          f" {mistyped_option._option_name} {chars.dot} {mistyped_argument}")
        for mistyped_option, mistyped_argument in mistyped_option_arguments_multiple:
            errors.append(f"Mistyped argument for option:"
                          f" {mistyped_option._option_name} {chars.dot} {mistyped_argument}")
        for violation in rule_violations_exactly_one_of_toomany:
            errors.append(f"Exactly one of these options may be specified: {', '.join(violation)}")
        for violation in rule_violations_exactly_one_of_missing:
            errors.append(f"Exactly one of these options must be specified: {', '.join(violation)}")
        for violation in rule_violations_at_least_one_of_missing:
            errors.append(f"At least one of these options must be specified: {', '.join(violation)}")
        for violation in rule_violations_at_most_one_of_toomany:
            errors.append(f"At most one of these options must be specified: {', '.join(violation)}")
        for violation in rule_violations_depends_on:
            errors.append(f"Option {violation[0]} depends on one of option(s): {violation[1]}")

        # Report any errors if desired.
        if (report is not False) and errors:
            if not callable(printf):
                printf = lambda *args, **kwargs: print(*args, **kwargs, file=sys.stderr)  # noqa
            for error in errors:
                printf(error)
            if exit is True:
                sys.exit(1)

        return errors

    def _process_option_definitions(self, *args) -> None:

        def flatten(*args):
            flattened_args = []
            def flatten(*args):  # noqa
                nonlocal flattened_args
                for arg in args:
                    if isinstance(arg, (list, tuple)):
                        for itemi in range(len(arg)):
                            if isinstance(item := arg[itemi], int) and ((itemi + 1) < len(arg)):
                                if isinstance(next_item := arg[itemi + 1], tuple) or isinstance(next_item, str):
                                    if (item & Argv.OPTIONAL) != Argv.OPTIONAL:
                                        item |= Argv.REQUIRED
                            flatten(item)
                    else:
                        flattened_args.append(arg)
            flatten(args)
            return flattened_args

        option_definitions = Argv._OptionDefinitions(fuzzy=self._fuzzy)

        if (len(args) == 1) and isinstance(options := args[0], dict):
            args = []
            copied_options = {}
            for option_type in options:
                if option_type == ARGV.REQUIRED:
                    copied_options[ARGV.REQUIRED(bool)] = options[option_type]
                else:
                    copied_options[option_type] = options[option_type]
            options = copied_options
            for option_type in options:
                option_options = options[option_type]
                if isinstance(option_type, str):
                    if (option_type.startswith(Argv._RULE_PREFIX) and
                        (option_type := option_type[Argv._RULE_PREFIX_LEN:]) and
                        ((index := option_type.find(Argv._RULE_DELIMITER)) > 0) and
                        (option_type := option_type[:index])):  # noqa
                        if option_type == "exactly_one_of":
                            option_definitions.add_rule_exactly_one_of(option_options)
                            continue
                        elif option_type == "at_least_one_of":
                            option_definitions.add_rule_at_least_one_of(option_options)
                            continue
                        elif option_type == "at_most_one_of":
                            option_definitions.add_rule_at_most_one_of(option_options)
                            continue
                        elif option_type == "depends_on":
                            option_definitions.add_rule_depends_on(option_options)
                            continue
                        option_type = to_integer(option_type[0:index])
                    elif option_type_components := to_string_list(option_type.split(Argv._RULE_DELIMITER)):
                        option_type = to_integer(option_type_components[0])
                        if option_default := option_type_components[1] if len(option_type_components) > 1 else None:
                            option_options.append(ARGV.DEFAULT)
                            option_options.append(option_default)

                default_value = None
                if isinstance(option_options, (list, tuple)) and option_options:
                    for optioni in range(len(option_options) - 1, -1, -1):
                        if option_options[optioni] == Argv.DEFAULT:
                            if (optioni + 1) < len(option_options):
                                default_value = option_options[optioni + 1]
                                option_options = option_options[0:optioni]
                                break
                if Argv._is_option_type(option_type) and (option_options := to_non_empty_string_list(option_options)):
                    if not any(self._is_option(option) for option in option_options):
                        if default_is_but_should_not_be_boolean := ((option_type & Argv.BOOLEAN) == Argv.BOOLEAN):
                            option_type &= ~Argv.BOOLEAN
                        if option_type & (Argv.STRINGS | Argv.INTEGERS | Argv.INTEGERS | Argv.DEFAULTS):
                            option_type |= Argv.DEFAULTS
                            if default_is_but_should_not_be_boolean:
                                option_type | ~Argv.STRING
                        else:
                            option_type |= Argv.DEFAULT
                            if default_is_but_should_not_be_boolean:
                                option_type | ~Argv.STRINGS
                    if default_value is not None:
                        option_options[0] = f"{option_options[0]}={str(default_value)}"
                    args.append(option_type)
                    args.append(option_options)

        if args := flatten(args):
            option_type = None ; options = [] ; parsing_options = None  # noqa
            for arg in args:
                if Argv._is_option_type(arg):
                    if (parsing_options is True) and option_type and options:
                        option_definitions.define_option(option_type, options)
                        option_type = None ; options = []  # noqa
                    parsing_options = False
                    option_type = arg
                elif isinstance(arg, str) and (arg := arg.strip()):
                    if (parsing_options is False) and option_type and options:
                        option_definitions.define_option(option_type, options)
                        option_type = None ; options = []  # noqa
                    parsing_options = True
                    options.append(arg)
            if options:
                option_definitions.define_option(option_type or Argv.BOOLEAN, options)
        return option_definitions

    def _defined_value_options(self) -> List[Argv._Option]:
        defined_value_options = []
        for option in self._option_definitions._definitions:
            if hasattr(self._values, option._property_name):
                defined_value_options.append(option)
        return defined_value_options

    def _find_options(self, options: Union[List[str], str]) -> List[Argv._Option]:
        found_options = []
        if options := to_non_empty_string_list(options):
            for option_definition in self._option_definitions._definitions:
                for option_definition_option in option_definition._options:
                    if self._is_option_any_of(option_definition_option, options):
                        found_options.append(option_definition)
                        break
        return found_options

    def _find_option(self, option: str) -> Optional[Argv._Option]:
        for option_definition in self._option_definitions._definitions:
            for option_definition_option in option_definition._options:
                if self._is_option_any_of(option_definition_option, option):
                    return option_definition
        return None

    def _is_option(self, value: str) -> bool:
        if self._escaping:
            return False
        if isinstance(value, str) and (value := value.strip()):
            if value.startswith(Argv._OPTION_PREFIX) and (value := value[Argv._OPTION_PREFIX_LEN:].strip()):
                return True
            elif self._fuzzy:
                if value.startswith(Argv._FUZZY_OPTION_PREFIX) and value[Argv._FUZZY_OPTION_PREFIX_LEN:].strip():
                    return True
        return False

    def _is_option_any_of(self, value: str, options: Union[List[str], str]) -> bool:
        def match(value: str, option: str) -> bool:
            if (isinstance(option, str) and (option := option.strip()) and
                isinstance(value, str) and (value := value.strip())):  # noqa
                if value == option:
                    return True
                elif (self._fuzzy and option.startswith(Argv._OPTION_PREFIX) and
                      (value == Argv._FUZZY_OPTION_PREFIX + option[Argv._OPTION_PREFIX_LEN:])):  # noqa
                    return True
            return False
        if isinstance(options, str):
            options = [options]
        return any(match(value, option) for option in options) if isinstance(options, list) else False

    @property
    def _property_names(self) -> List[str]:
        property_names = []
        for option in self._option_definitions._definitions:
            property_names.append(option._property_name)
        return property_names

    @staticmethod
    def _property_name_ize(value: str) -> str:
        return value.replace("-", "_") if isinstance(value, str) else ""

    @property
    def _dict(self) -> dict:
        values = {}
        for property_name in self._property_names:
            values[property_name] = getattr(self._values, property_name, None)
        return sort_dictionary(values)

    @staticmethod
    def _is_option_type(option_type: Any) -> bool:
        if isinstance(option_type, int):
            return (((Argv.BOOLEAN | Argv.DEFAULT | Argv.DEFAULTS |
                      Argv.FLOAT | Argv.FLOATS | Argv.INTEGER | Argv.INTEGERS |
                      Argv.STRING | Argv.STRINGS | Argv.OPTIONAL | Argv.REQUIRED) & option_type) == option_type)
        return False

    @property
    def _peek(self) -> Optional[str]:
        return Argv._Arg(self._argv[self._argi], self) if self._argi < len(self._argv) else Argv._Arg(None, self)

    @property
    def _next(self) -> Optional[str]:
        if (value := self._peek) is not None:
            if self._delete:
                del self._argv[0]
            else:
                self._argi += 1
            if (not self._escaping) and (value == Argv._ESCAPE_OPTION):
                self._escaping = True
                return self._next
            return value
        return None

    def __iter__(self) -> Argv:
        return self

    def __next__(self) -> Optional[str]:
        if self._argi >= len(self._argv):
            raise StopIteration
        arg = self._argv[self._argi]
        if self._delete:
            del self._argv[0]
        else:
            self._argi += 1
        if (arg == Argv._ESCAPE_OPTION) and self._escape and (not self._escaping):
            self._escaping = True
            return self.__next__()
        return Argv._Arg(arg, self)

    def __getattr__(self, name: str):
        if not hasattr(self._values, name):
            raise AttributeError(f"Property for argument not found: {name}")
        return getattr(self._values, name)


class ARGV(Argv):

    @staticmethod
    def OPTIONAL(type: Optional[Type[Union[str, int, float, bool]]] = None,
                 default: Any = None, required: bool = False) -> str:
        if isinstance(type, list) and (len(type) == 1):
            type = type[0]
            if type == str: option_type = Argv.STRINGS  # noqa
            elif type == int: option_type = Argv.INTEGERS  # noqa
            elif type == float: option_type = Argv.FLOATS  # noqa
            else:
                raise Exception("Invalid argument to ARGV.OPTIONAL")
        else:
            if type == str: option_type = Argv.STRING  # noqa
            elif type == int: option_type = Argv.INTEGER  # noqa
            elif type == float: option_type = Argv.FLOAT  # noqa
            elif type == bool: option_type = Argv.BOOLEAN  # noqa
            else:
                raise Exception("Invalid argument to ARGV.OPTIONAL")
        if required is True:
            option_type |= Argv.REQUIRED
        else:
            option_type |= Argv.OPTIONAL
        if default is not None:
            return str(option_type) + Argv._RULE_DELIMITER + str(default) + Argv._RULE_DELIMITER + str(uuid())
        else:
            return str(option_type) + Argv._RULE_DELIMITER + Argv._RULE_DELIMITER + str(uuid())
        return str(option_type) + Argv._RULE_DELIMITER + str(uuid())

    @staticmethod
    def REQUIRED(type: Optional[Type[Union[str, int, float, bool]]] = bool, default: Any = None):
        return ARGV.OPTIONAL(type=type, default=default, required=True)

    @classmethod
    @property
    def EXACTLY_ONE_OF(cls):
        return f"{Argv._RULE_PREFIX}exactly_one_of{Argv._RULE_DELIMITER}{str(uuid())}"

    @classmethod
    @property
    def AT_LEAST_ONE_OF(cls):
        return f"{Argv._RULE_PREFIX}at_least_one_of{Argv._RULE_DELIMITER}{str(uuid())}"

    @classmethod
    @property
    def AT_MOST_ONE_OF(cls):
        return f"{Argv._RULE_PREFIX}at_most_one_of{Argv._RULE_DELIMITER}{str(uuid())}"

    @classmethod
    @property
    def DEPENDENCY(cls, *args):
        return f"{Argv._RULE_PREFIX}depends_on:{str(uuid())}"

    @classmethod
    @property
    def DEPENDS_ON(cls, *args):
        return f"{Argv._RULE_PREFIX}depends_on:{str(uuid())}"

    @classmethod
    @property
    def ALLOW_ONLY(cls, *args):  # TODO
        return f"{Argv._RULE_PREFIX}allow_only:{str(uuid())}"


DEFAULT = ARGV.DEFAULT
OPTIONAL = ARGV.OPTIONAL
REQUIRED = ARGV.REQUIRED
AT_MOST_ONE_OF = ARGV.AT_MOST_ONE_OF
AT_LEAST_ONE_OF = ARGV.AT_LEAST_ONE_OF
EXACTLY_ONE_OF = ARGV.EXACTLY_ONE_OF
DEPENDENCY = ARGV.DEPENDENCY
DEPENDS_ON = ARGV.DEPENDS_ON
ALLOW_ONLY = ARGV.ALLOW_ONLY



from typing import Callable, Optional
from hms_utils.config.config_with_secrets import ConfigWithSecrets
from hms_utils.config.config_with_aws_macros import ConfigWithAwsMacros
from hms_utils.dictionary_parented import DictionaryParented as JSON


class Config(ConfigWithSecrets, ConfigWithAwsMacros):

    def __init__(self, config: JSON,
                 name: Optional[str] = None,
                 path_separator: Optional[str] = None,
                 custom_macro_lookup: Optional[Callable] = None,
                 raise_exception: bool = False,
                 secrets: bool = False,
                 decrypted: bool = False,
                 noaws: bool = False, **kwargs) -> None:

        super().__init__(config=config,
                         name=name,
                         path_separator=path_separator,
                         custom_macro_lookup=custom_macro_lookup,
                         raise_exception=raise_exception,
                         secrets=secrets,
                         decrypted=decrypted,
                         noaws=noaws, **kwargs)



from __future__ import annotations
from copy import deepcopy
from typing import Any, Callable, List, Optional, Tuple, Union
from hms_utils.dictionary_utils import sort_dictionary
from hms_utils.path_utils import unpack_path
from hms_utils.type_utils import is_primitive_type


# This JSON class isa dictionary type which also suport "parent" property for each/every sub-dictionary
# within the main dictionary. Should be able to use EXACTLY like a dicttionary type after creating with
# JSON(dict); including copying and setting properties to any type including either other dictionaies
# or JSON objects. Also has a "root" property on each sub-dictionary referring to the main/root one;
# this just walks up the parent properties to the top (where parent is of course None).
#
class JSON(dict):

    _PATH_SEPARATOR = "/"

    def __init__(self, data: Optional[Union[dict, JSON]] = None, rvalue: Optional[Callable] = None) -> None:
        if isinstance(data, JSON):
            data = dict(data)
        elif not isinstance(data, dict):
            data = {}
        super().__init__(data)
        self._initialize(rvalue)

    def _initialize(self, rvalue: Optional[Callable] = None) -> None:
        if not callable(rvalue):
            rvalue = None
        self._parent = None
        for key in self:
            value = super(JSON, self).__getitem__(key)
            if isinstance(value, dict):
                value = JSON(value, rvalue=rvalue)
                value._parent = self
                super(JSON, self).__setitem__(key, value)
            elif isinstance(value, list):
                value_list = []
                for element in value:
                    if isinstance(element, dict):
                        value_list.append(JSON(element, rvalue=rvalue))
                    else:
                        value_list.append(element)
                super().__setitem__(key, value_list)
            elif rvalue and is_primitive_type(value):
                super().__setitem__(key, rvalue(value))

    @property
    def parent(self) -> Optional[JSON]:
        return self._parent

    @property
    def root(self) -> Optional[JSON]:
        node = self
        while True:
            if node._parent is None:
                return node
            node = node._parent

    def context_path(self, path_separator: Optional[Union[str, bool]] = None,
                     path_rooted: bool = False, path_suffix: Optional[str] = None) -> Union[List[str], str]:
        # FYI we only actually use this in hms_config for diagnostic messages.
        context = self
        context_path = []
        context_parent = context._parent
        while context_parent:
            for key in context_parent:
                if context_parent[key] == context:
                    context_path.insert(0, key)
            context = context._parent
            context_parent = context_parent._parent
        if isinstance(path_suffix, str) and path_suffix:
            context_path.append(path_suffix)
        if path_separator is True:
            path_separator = JSON._PATH_SEPARATOR
        elif (path_separator is False) or (not isinstance(path_separator, str)):
            path_separator = None
        if path_separator:
            if path_rooted is True:
                return path_separator + path_separator.join(context_path)
            return path_separator.join(context_path)
        return context_path

    @property
    def path(self) -> str:
        return self.context_path(path_separator=True, path_rooted=id(self) == id(self.root))

    def sorted(self, reverse: bool = False) -> JSON:
        return JSON(sort_dictionary(self.root, reverse=reverse)).lookup(self.path)

    def merge(self, secondary: JSON, path_separator: Optional[str] = None) -> Tuple[JSON, List[str], List[str]]:
        # Merges the given secondary JSON object into a COPY of this JSON object; but does not overwrite
        # anything in this JSON object; anything that would otherwise overwrite is ignored. Returns a tuple
        # with (left-to-right) the (new) merged dictionary, a list of paths which were actually merged from the
        # secondary, and a list of paths which were not merged from the secondary, i.e. because they would have
        # overwritten that item in the copy of this JSON object; path delimiter is the given path_separator.
        def merge(primary: JSON, secondary: JSON) -> Tuple[JSON, List[str], List[str]]:
            nonlocal path_separator
            if not (isinstance(primary, dict) or isinstance(secondary, dict)):
                return None, None, None
            merged = deepcopy(primary) ; merged_paths = [] ; unmerged_paths = []  # noqa
            def merge(primary: dict, secondary: dict, path: str = "") -> None:  # noqa
                nonlocal unmerged_paths, path_separator
                for key, value in secondary.items():
                    key_path = f"{path}{path_separator}{key}" if path else key
                    if key not in primary:
                        primary[key] = secondary[key]
                        merged_paths.append(key_path)
                    elif isinstance(primary[key], dict) and isinstance(secondary[key], dict):
                        merge(primary[key], secondary[key], path=key_path)
                    else:
                        unmerged_paths.append(key_path)
            merge(merged, secondary)
            return JSON(merged), merged_paths, unmerged_paths
        if (not isinstance(path_separator, str)) or (not path_separator):
            path_separator = JSON._PATH_SEPARATOR
        return merge(self, secondary)

    def lookup(data, path: str, path_separator: Optional[str] = None) -> Optional[Union[Any, JSON]]:
        if (not isinstance(path, str)) or (not path):
            return None, data
        if (not isinstance(path_separator, str)) or (not path_separator):
            path_separator = JSON._PATH_SEPARATOR
        if not (path_components := unpack_path(path, path_separator=path_separator)):
            return None, data
        if path_components[0] == path_separator:
            if not (path_components := path_components[1:]):
                return data
            context = data.root
        else:
            context = data
        value = None
        for path_component_index, path_component in enumerate(path_components):
            if (value := context.get(path_component)) is None:
                break
            if isinstance(value, JSON):
                context = value
            elif path_component_index < (len(path_components) - 1):
                value = None
                break
        return value

    def duplicate(self, rvalue: Optional[Callable] = None) -> JSON:
        return JSON(self.root, rvalue=rvalue).lookup(self.path)

    def __setitem__(self, key: Any, value: Any) -> None:
        if isinstance(value, dict):
            if not isinstance(value, JSON):
                value = JSON(value)
            if id(value._parent) != id(self):
                if isinstance(value, JSON):
                    copied_value = deepcopy(value)
                    value = copied_value
                else:
                    value = JSON(value)
                value._parent = self
        super().__setitem__(key, value)

    def __deepcopy__(self, memo) -> JSON:
        return JSON(deepcopy(dict(self), memo))


DictionaryParented = JSON



import io
import json
import os

# hms-portal-read --env fourfront-data /files-processed --limit 1000 --inserts --output exported-from-fourfront-data-files-processed-limit-1000-20241121.json  # noqa
file = os.path.expanduser("exported-from-fourfront-data-files-processed-limit-1000-20241121.json")
file = os.path.expanduser("exported-from-fourfront-data-files-processed-limit-10000-20241121.json")
file = os.path.expanduser("~/repos/etc/repo-extras/portal/4dn/data/exported-from-fourfront-data-experiment-set-replicates-4DNESXP5VE8C-20241120-3-noignore-static-content/file_processed.json")  # noqa
md5sums = {}
md5sums_duplicates = {}
with io.open(file, "r") as f:
    items = json.load(f)
    if isinstance(items, list):
        items = {"Dummy": items}
    for item_type in items:
        for item in items[item_type]:
            if md5sum := item.get("md5sum"):
                item_uuid = item.get("uuid")
                if md5sum_uuids := md5sums.get(md5sum):
                    print(f"DUPLICATE MD5: {md5sum} | uuid: {item_uuid} | {md5sum_uuids}")
                    md5sum_uuids.append(item_uuid)
                    md5sums_duplicates[md5sum] = md5sum_uuids
                else:
                    md5sums[md5sum] = [item_uuid]
print(f"TOTAL DUPLICATE MD5: {len(md5sums_duplicates)}")



from __future__ import annotations
from boto3 import client as BotoClient
from functools import lru_cache
import re
import time
from typing import Dict, Generator, List, Optional, Tuple
from dcicutils.datetime_utils import format_datetime
from hms_utils.github_utils import get_github_commit_date, get_github_latest_commit


def get_aws_ecr_build_info(image_repo_or_arn: str, image_tag: str,
                           image_digest: str, previous_builds: int = 8) -> Optional[dict]:
    """
    Returns a dictionary with info about the three most recent CodeBuild builds
    for the given image repo and tag, or None if none found.
    """
    if not image_tag:
        image_repo, image_tag = _get_image_repo_and_tag(image_repo_or_arn)
    else:
        image_repo = image_repo_or_arn

    codebuild = BotoClient("codebuild")

    def get_projects() -> List[str]:

        projects = codebuild.list_projects()["projects"]

        # If there is a project with the same name as the image_repo then look at that one first,
        # or secondarily, prefer projects whose names contain the image_repo and/or image_tag;
        # we do this just for performance, to try to reduce the number of boto calls we make.

        def prefer_project(preferred_project: str):
            projects.remove(preferred_project)
            return [preferred_project, *projects]

        projects = [project for project in projects
                    if "pipeline" not in project.lower() and "tibanna" not in project.lower()]

        preferred_project = [project for project in projects if project == image_repo]
        if preferred_project:
            return prefer_project(preferred_project[0])
        preferred_project = [project for project in projects
                             if image_repo.lower() in project.lower() and image_tag.lower() in project.lower()]
        if preferred_project:
            return prefer_project(preferred_project[0])
        preferred_project = [project for project in projects if image_repo.lower() in project.lower()]
        if preferred_project:
            return prefer_project(preferred_project[0])
        preferred_project = [project for project in projects if image_tag.lower() in project.lower()]
        if preferred_project:
            return prefer_project(preferred_project[0])
        return projects

    def get_relevant_builds(project: str) -> Generator[Optional[Dict], None, None]:

        # For efficiency, and the most common actual case, get builds three at a time; i.e. since we
        # want to the three most recent (relevant) builds, and they are usually together at the start
        # of the (list_build_for_projects) list ordered (descending) by build (creation) time;
        # but of course, just in case, we need to handle the general case.

        def get_relevant_build_info(build: Optional[Dict]) -> Optional[Dict]:

            def create_build_info(build: Dict) -> Dict:
                return {
                    # "arn": _shortened_arn(build["arn"]),
                    "arn": build["arn"],
                    "project": project,
                    "image_repo": image_repo,
                    "image_tag": image_tag,
                    "github": build.get("source", {}).get("location"),
                    "branch": build.get("sourceVersion"),
                    "commit": build.get("resolvedSourceVersion"),
                    "number": build.get("buildNumber"),
                    # "initiator": _shortened_arn(build.get("initiator")),
                    "initiator": build.get("initiator"),
                    "status": build.get("buildStatus"),
                    "success": is_build_success(build),
                    "finished": build.get("buildComplete"),
                    "started_at": format_datetime(build.get("startTime")),
                    "finished_at": format_datetime(build.get("endTime")),
                    "log_group": build.get("logs", {}).get("groupName"),
                    "log_stream": build.get("logs", {}).get("streamName")
                }

            def is_build_success(build: Dict) -> bool:
                return build.get("buildStatus", "").upper() in ["SUCCEEDED", "SUCCESS"]

            def find_environment_variable(environment_variables: List[Dict], name: str) -> Optional[str]:
                value = [item["value"] for item in environment_variables if item["name"] == name]
                return value[0] if len(value) == 1 else None

            if build and is_build_success(build):
                environment_variables = build.get("environment", {}).get("environmentVariables", {})
                build_image_repo = find_environment_variable(environment_variables, "IMAGE_REPO_NAME")
                if build_image_repo:
                    pass  # TODO
                build_image_tag = find_environment_variable(environment_variables, "IMAGE_TAG")
                if build_image_repo == image_repo and build_image_tag == image_tag:
                    return create_build_info(build)
            return None

        get_builds_from_boto_this_many_at_a_time = 4
        nbuilds = 0
        next_token = None
        while True:
            if nbuilds >= previous_builds:
                break
            if next_token:
                build_ids = codebuild.list_builds_for_project(projectName=project, nextToken=next_token)
            else:
                build_ids = codebuild.list_builds_for_project(projectName=project, sortOrder="DESCENDING")
            next_token = build_ids.get("nextToken")
            build_ids = build_ids["ids"]
            while build_ids:
                build_ids_batch = build_ids[:get_builds_from_boto_this_many_at_a_time]
                build_ids = build_ids[get_builds_from_boto_this_many_at_a_time:]
                builds = codebuild.batch_get_builds(ids=build_ids_batch)["builds"]
                for build in builds:
                    if nbuilds >= previous_builds:
                        break
                    nbuilds += 1
                    build = get_relevant_build_info(build)
                    if build:
                        yield build
            if not next_token:
                break

    for project in get_projects():
        for build in get_relevant_builds(project):
            if image_digest:
                if build_image_digest := _get_aws_codebuild_digest(build["log_group"], build["log_stream"], image_tag):
                    if image_digest == build_image_digest:
                        return build
    return None


@lru_cache
def get_build_info(image_repo: str, image_tag: str, image_digest: Optional[str] = None) -> Optional[Dict]:
    # Cache this result within the enclosing function; for the below services loop.
    return get_aws_ecr_build_info(image_repo, image_tag, image_digest)


def get_image_build_info(image_repo: str, image_tag: str, image_digest: str) -> Optional[dict]:
    if build_info := get_build_info(image_repo, image_tag, image_digest):
        if git_repo := build_info.get("github"):
            git_repo = git_repo.replace("https://github.com/", "")
        git_branch = build_info.get("branch")
        git_commit = build_info.get("commit")
        git_commit_date = get_github_commit_date(git_repo, git_commit)
        git_commit_latest = get_github_latest_commit(git_repo, git_branch)
        if git_commit_latest != git_commit:
            git_commit_latest_date = get_github_commit_date(git_repo, git_commit_latest)
        else:
            git_commit_latest_date = git_commit_date
        return {
           "build_project": build_info["project"],
           "repo": git_repo,
           "branch": git_branch,
           "commit": git_commit,
           "commit_date": git_commit_date,
           "commit_latest": git_commit_latest,
           "commit_latest_date": git_commit_latest_date
        }
    return None


def _get_image_repo_and_tag(image_arn: str) -> Tuple[Optional[str], Optional[str]]:
    # image_arn = _shortened_arn(image_arn)
    image_arn = image_arn
    parts = image_arn.split(":")
    return (parts[0], parts[1]) if len(parts) == 2 else (None, None)


def _get_aws_codebuild_digest(log_group: str, log_stream: str, image_tag: Optional[str] = None) -> Optional[str]:
    logs = BotoClient("logs")
    sha256_pattern = re.compile(r"sha256:([0-9a-f]{64})")
    # For some reason this (rarely-ish) intermittently fails with no error;
    # the results just do not contain the digest; don't know why so try a few (4) times.
    for n in range(4):
        if n > 1:
            time.sleep(0.05)
        if not (log_events := logs.get_log_events(logGroupName=log_group,
                                                  logStreamName=log_stream, startFromHead=False)["events"]):
            log_events = logs.get_log_events(logGroupName=log_group,
                                             logStreamName=log_stream, startFromHead=True)["events"]
        for log_event in log_events:
            msg = log_event.get("message")
            # The entrypoint_deployment.bash script at least partially
            # creates this log output, which includes a line like this:
            # green: digest: sha256:c1204f9ff576105d9a56828e2c0645cc6dbcf91abca767ef6fe033a60c483f10 size: 7632
            if msg and "digest:" in msg and "size:" in msg and (not image_tag or f"{image_tag}:" in msg):
                match = sha256_pattern.search(msg)
                if match:
                    return "sha256:" + match.group(1)
    return None



import os
from contextlib import contextmanager


@contextmanager
def os_environ(key: str, value: str):
    saved_value = os.environ.get(key)
    try:
        os.environ[key] = value
        yield
    finally:
        if saved_value is not None:
            os.environ[key] = saved_value
        else:
            del os.environ[key]



import concurrent.futures
from typing import Callable, Iterable


def run_concurrently(functions: Iterable[Callable], nthreads: int = 4) -> None:
    # FYI: Not pulling in from dcicutils.misc_utils becausethere is
    # a call to logging.basicConfig() which is (for some reason) causing
    # exceptions within the asynchronous function calls to be output.
    if not (isinstance(nthreads, int) and (nthreads >= 0)):
        nthreads = 4
    if nthreads == 0:
        for function in functions:
            function()
        return
    with concurrent.futures.ThreadPoolExecutor(max_workers=nthreads) as executor:
        concurrent.futures.as_completed([executor.submit(f) for f in functions])



import requests
import sys
from typing import Optional, Tuple
from datetime import datetime
from dcicutils.captured_output import captured_output
from dcicutils.datetime_utils import format_datetime
from hms_utils.chars import chars
from hms_utils.config.config import Config
from hms_utils.github_utils import get_github_commit_date, get_github_latest_commit

ENVS = [
    {
        "name": "smaht-wolf",
        "url": "https://wolf.smaht.org"
    },
    {
        "name": "smaht-devtest",
        "url": "https://devtest.smaht.org"
    },
    {
        "name": "smaht-data",
        "url": "https://data.smaht.org",
        "blue_green": True,
        "repo": "smaht-dac/smaht-portal",
        "repo_branch": "main"
    },
    {
        "name": "smaht-staging",
        "url": "https://staging.smaht.org",
        "blue_green": True,
        "repo": "smaht-dac/smaht-portal",
        "repo_branch": "main"
    },
    {
        "name": "cgap-devtest",
        "url": "https://cgap-devtest.hms.harvard.edu",
        "defunct": True
    },
    {
        "name": "cgap-dbmi",
        "url": "https://cgap-dbmi.hms.harvard.edu"
    },
    {
        "name": "cgap-dfci",
        "url": "https://cgap-dfci.hms.harvard.edu",
        "defunct": True
    },
    {
        "name": "cgap-mgb",
        "url": "https://cgap-mgb.hms.harvard.edu"
    },
    {
        "name": "cgap-msa",
        "url": "https://cgap-msa.hms.harvard.edu"
    },
    {
        "name": "cgap-training",
        "url": "https://cgap-training.hms.harvard.edu"
    },
    {
        "name": "cgap-wolf",
        "url": "https://cgap-wolf.hms.harvard.edu"
    },
    {
        "name": "4dn-data",
        "url": "https://data.4dnucleome.org",
        "blue_green": True
    },
    {
        "name": "4dn-staging",
        "url": "https://staging.4dnucleome.org",
        "blue_green": True
    },
    {
        "name": "4dn-mastertest",
        "url": "https://mastertest.4dnucleome.org"
    },
    {
        "name": "4dn-hotseat",
        "url": "http://fourfront-hotseat-1650461284.us-east-1.elb.amazonaws.com"
    },
    {
        "name": "4dn-webdev",
        "url": "http://fourfront-webdev-1726959209.us-east-1.elb.amazonaws.com"
    }
]


def usage():
    print("usage: envs [--health] [pattern]")
    exit(1)


def main():

    env_pattern = None
    show_health = False
    health_key_pattern = None
    health_value_pattern = None
    verbose = False

    argi = 0
    while argi < len(argv := sys.argv[1:]):
        arg = argv[argi]
        if (arg == "--health") or (arg == "-health") or (arg == "health"):
            show_health = True
        elif arg.startswith("--health:") or arg.startswith("--health:") or arg.startswith("health:"):
            health_key_pattern = arg[arg.find(":") + 1:]
        elif arg.startswith("--health/") or arg.startswith("--health/") or arg.startswith("health/"):
            health_value_pattern = arg[arg.find("/") + 1:]
        elif arg in ["--verbose", "-verbose"]:
            verbose = True
        elif arg.startswith("-"):
            usage()
        else:
            env_pattern = arg
        argi += 1

    if verbose:
        github_token = Config.default().lookup("GITHUB_TOKEN", show=True)

    envs = [env for env in ENVS if env.get("defunct") is not True]
    if env_pattern:
        envs = [env for env in envs if env_pattern.lower() in env.get("name").lower()]
    for env in sorted(envs, key=lambda env: env.get("name")):
        name = env["name"]
        if url := env.get("url"):
            health, issue = get_health(url)
            if not health:
                print(f"{name}: Cannot connect: {url}")
                continue
            version = health.get("project_version")
            blue_green = None
            if env.get("blue_green"):
                if blue_green_identification := health.get("beanstalk_env", "").lower():
                    if "blue" in blue_green_identification:
                        blue_green = "blue"
                    elif "green" in blue_green_identification:
                        blue_green = "green"
            print(f"{chars.rarrow} {name}: {version}"
                  f"{f' ({blue_green})' if blue_green else ''}{f' ({issue})' if issue else ''}")
            if show_health or health_key_pattern or health_value_pattern:
                health = {key: health[key] for key in sorted(health)}
                for key in health:
                    if key.startswith("@"):
                        continue
                    if not isinstance(value := health[key], str):
                        value = str(value)
                    if health_key_pattern:
                        if health_key_pattern.lower() not in key.lower():
                            continue
                    if health_value_pattern:
                        if health_value_pattern.lower() not in value.lower():
                            continue
                    print(f"  - {key}: {value}")
            if verbose:
                if ((repo := env.get("repo")) and
                    (health_git := health.get("git")) and (health_git_commit := health_git.get("commit"))):  # noqa
                    health_git_commit_date = get_github_commit_date(repo, health_git_commit, github_token=github_token)
                    if latest_repo_commit := get_github_latest_commit(repo, env.get("repo_branch", "master"),
                                                                      github_token=github_token):
                        latest_repo_commit_date = get_github_commit_date(repo, latest_repo_commit,
                                                                         github_token=github_token)
                    howold = get_duration(latest_repo_commit_date, health_git_commit_date)
                    print(f"  - git: {health_git_commit} {chars.dot}"
                          f" {format_datetime(health_git_commit_date)} ({howold})")
                    pass


def get_health(url: str) -> Tuple[Optional[dict], Optional[str]]:
    with captured_output(capture=True):
        try:
            if ((response := requests.get(f"{url}/health?format=json", verify=True)) and
                (response.status_code == 200) and
                (response := response.json())):  # noqa
                return response, None
        except Exception:
            if ((response := requests.get(f"{url}/health?format=json", verify=False)) and
                (response.status_code == 200) and
                (response := response.json())):  # noqa
                return response, "certificate issue"
            pass
    return None


def get_duration(more_recent, less_recent):
    delta_seconds = (datetime.fromisoformat(more_recent) - datetime.fromisoformat(less_recent)).total_seconds()
    if delta_seconds > 0:
        if (delta_hours := int(delta_seconds / 3600)) > 0:
            if (delta_days := int(delta_hours / 24)) > 0:
                return f"-{delta_days}d"
            else:
                return f"-{delta_hours}h"


if __name__ == "__main__":
    status = main()
    sys.exit(status if isinstance(status, int) else 0)



from typing import Any, Callable, List, Optional, Tuple, Union
import iterm2
from iterm2.session import Session
from iterm2.connection import Connection
import time


def iterm_run(function: Callable) -> None:
    async def main(connection):
        session = await iterm_init_session(connection)
        await function(session)
    iterm2.run_until_complete(main)


async def iterm_init_session(connection: Connection) -> Session:
    return (await iterm_init(connection))[3]


async def iterm_init(connection: Connection):
    app = await iterm2.async_get_app(connection)
    tab = (window := app.current_window).current_tab
    return app, window, tab, tab.current_session


async def new_pane(session: Session,
                   commands: Union[List[str], str],
                   title: Optional[str] = None,
                   on_output: Optional[Tuple[str, Callable]] = None) -> Session:
    pane = await session.async_split_pane(vertical=False)
    if isinstance(title, str) and title:
        await pane.async_send_text(f"echo -e \"\\033];{title}\007\"\n")
    await pane_output(pane, commands)
    if on_output and len(on_output) > 1:
        await monitor_pane(pane, trigger=on_output[0], callback=on_output[1],
                           data=on_output[3] if len(on_output) > 2 else session)
    return pane


async def pane_output(pane: Session, commands: Union[List[str], str]) -> None:
    if isinstance(commands, str) and commands:
        commands = [commands]
    if isinstance(commands, list):
        for command in commands:
            if command:
                if not command.endswith("\n"):
                    command = command + "\n"
                await pane.async_send_text(command)


async def monitor_pane(pane: Session, trigger: str, callback: Callable, data: Any) -> None:
    line_number = 0
    time.sleep(2)
    while True:
        time.sleep(1)
        lines = [line.string for line in
                 await pane.async_get_contents(first_line=line_number, number_of_lines=25) if line.string]
        line_number += len(lines)
        for line in lines:
            if trigger in line:
                return await callback(data)



import json
import sys
from typing import List
from hms_utils.argv import ARGV
from hms_utils.portal.portal_utils import Portal as Portal
from hms_utils.type_utils import is_uuid


def main():

    argv = ARGV({
        ARGV.OPTIONAL(str): ["--env"],
        ARGV.REQUIRED(str): ["query"],
        ARGV.OPTIONAL([str]): ["--consortia", "--consortium", "--c"],
        ARGV.OPTIONAL([str]): ["--groups", "--group", "--g"],
        ARGV.OPTIONAL([str]): ["--submission-centers", "--submission-center", "--sc", "--s"],
        ARGV.OPTIONAL(bool): ["--verbose"],
        ARGV.OPTIONAL(bool): ["--debug"],
        ARGV.OPTIONAL(bool): ["--ping"]
    })

    portal = Portal.create(argv.env, verbose=argv.verbose, debug=argv.debug, ping=argv.ping)

    query = argv.query
    if not (item := portal.get_metadata(query)):
        print("Cannot find item: {argv.query")

    if argv.consortia:
        if (len(argv.consortia) == 1) and (argv.consortia[0].lower() in ["none", "null", "no", "empty"]):
            if argv.verbose:
                print(f"Removing consortia for {argv.query}")
            portal.delete_metadata_property(query, "consortia")
        else:
            if (consortia := _get_consortia(portal, argv.consortia)):
                if argv.verbose:
                    print(f"Setting consortia for {argv.query} to: {', '.join(consortia)}")
                portal.patch_metadata(query, {"consortia": consortia})

    if argv.submission_centers:
        if (len(argv.submission_centers) == 1) and (argv.submission_centers[0] in ["none", "null", "no", "empty"]):
            # Deleting submission_centers does not seem to work; set it to empty list instead.
            # portal.delete_metadata_property(query, "submission_centers")
            if argv.verbose:
                print(f"Removing submission-centers for {argv.query}")
            portal.patch_metadata(query, {"submission_centers": []})
        else:
            if (submission_centers := _get_submission_centers(portal, argv.submission_centers)):
                if argv.verbose:
                    print(f"Setting submission-centers for {argv.query} to: {', '.join(submission_centers)}")
                portal.patch_metadata(query, {"submission_centers": submission_centers})

    if argv.groups:
        if (len(argv.groups) == 1) and (argv.groups[0] in ["none", "null", "no", "empty"]):
            portal.delete_metadata_property(query, "groups")
        else:
            if argv.verbose:
                print(f"Setting groups for {argv.query} to: {', '.join(argv.groups)}")
            portal.patch_metadata(query, {"groups": argv.groups})

    item = portal.get_metadata(query, raw=True, database=True)
    print(json.dumps(item, indent=4))


def _get_consortia(portal: Portal, values: List[str]) -> List[str]:
    return _get_affiliations(portal, values, "consortia")


def _get_submission_centers(portal: Portal, values: List[str]) -> List[str]:
    return _get_affiliations(portal, values, "submission-centers")


def _get_affiliations(portal: Portal, values: List[str], affiliation_name: str) -> List[str]:
    values_found = []
    if isinstance(values, list):
        for value in values:
            if isinstance(value, str):
                if ((value_found := portal.get_metadata(f"/{affiliation_name}/{value}", raise_exception=False)) and
                    is_uuid(value_found_uuid := value_found.get("uuid"))):  # noqa
                    values_found.append(value_found_uuid)
                else:
                    print(f"WARNING: Specified value ({affiliation_name}) not found: {value}", file=sys.stderr)
    return values_found


if __name__ == "__main__":
    main()



from __future__ import annotations
from enum import Enum, auto as enum_auto
from functools import lru_cache
import hashlib
import io
import json
import os
import re
import requests
import sys
import time
from typing import Any, Callable, List, Optional, Set, Tuple, Union
from uuid import uuid4
import yaml
from dcicutils.command_utils import yes_or_no
from dcicutils.misc_utils import to_snake_case
from hms_utils.argv import ARGV
from hms_utils.chars import chars
from hms_utils.datetime_utils import format_duration
from hms_utils.dictionary_utils import contains_uuid, delete_properties_from_dictionaries, find_dictionary_item
from hms_utils.dictionary_utils import get_property, get_uuids, get_referenced_uuids, sort_dictionary
from hms_utils.portal.portal_utils import Portal as PortalFromUtils
from hms_utils.threading_utils import run_concurrently
from hms_utils.type_utils import is_uuid
from hms_utils.version_utils import get_version

_ITEM_SID_PROPERTY_NAME = "sid"
_ITEM_UUID_PROPERTY_NAME = "uuid"
_ITEM_TYPE_PSEUDO_PROPERTY_NAME = "@@@__TYPE__@@@"
# _ITEM_IGNORE_REF_PROPERTIES = ["viewconfig", "higlass_uid", "blob_id", "static_content"]
_ITEM_IGNORE_REF_PROPERTIES = ["viewconfig", "higlass_uid", "blob_id"]  # "static_content"
_ITEM_MD5SUM_PROPERTY_NAME = "md5sum"

_ITEM_IGNORE_PROPERTIES_INSERTS = [
    "date_created",
    "last_modified",
    "schema_version",
    "submitted_by"
]

_ITEM_IGNORE_PROPERTIES_NOCRUFT = [
    *_ITEM_IGNORE_PROPERTIES_INSERTS,
    "principals_allowed",
    "actions"
]


class Portal(PortalFromUtils):

    class Access(Enum):
        OK = enum_auto()
        NOT_FOUND = enum_auto()
        NO_ACCESS = enum_auto()
        ERROR = enum_auto()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._get_call_count = 0
        self._get_metadata_call_count = 0
        self._get_call_duration = 0
        self._get_metadata_call_duration = 0
        self._raise_exception = kwargs.get("raise_exception") is True
        self._ignore_properties = []

    @property
    def get_call_count(self) -> int:
        return self._get_call_count

    @property
    def get_metadata_call_count(self) -> int:
        return self._get_metadata_call_count

    @property
    def get_call_duration(self) -> float:
        return self._get_call_duration

    @property
    def get_metadata_call_duration(self) -> float:
        return self._get_metadata_call_duration

    @property
    def ignore_properties(self) -> List[str]:
        return self._ignore_properties

    @ignore_properties.setter
    def ignore_properties(self, value: List[str]) -> None:
        self._ignore_properties = value if isinstance(value, list) else []

    def GET(self, query: str, metadata: bool = False, raw: bool = False, database: bool = False,
            limit: Optional[int] = None, offset: Optional[int] = None,
            field: Optional[str] = None, deleted: bool = False,
            raise_exception: bool = False) -> Optional[Union[List[dict], dict]]:
        items = None
        try:
            _debug(f"portal.get{'_metadata' if metadata else ''}: {query}"
                   f"{f' {chars.dot} raw' if raw else ''}"
                   f"{f' {chars.dot} database' if database else ''}"
                   f"{f' {chars.dot} limit: {limit}' if isinstance(limit, int) else ''}"
                   f"{f' {chars.dot} offset: {offset}' if isinstance(offset, int) else ''}"
                   f"{f' {chars.dot} field: {field}' if field else ''}"
                   f"{f' {chars.dot} deleted' if deleted else ''}")
            if metadata:
                self._get_metadata_call_count += 1
                started = time.time()
                items = self.get_metadata(query, raw=raw, database=database,
                                          limit=limit, offset=offset, deleted=deleted, field=field)
                self._get_metadata_call_duration += time.time() - started
            else:
                self._get_call_count += 1
                started = time.time()
                items = self.get(query, raw=raw, database=database,
                                 limit=limit, offset=offset, deleted=deleted, field=field)
                if items.status_code == 404:
                    return Portal.Access.NOT_FOUND
                elif items.status_code == 403:
                    return Portal.Access.NO_ACCESS
                items = items.json()
                self._get_call_duration += time.time() - started
        except Exception as e:
            if (raise_exception is True) or self._raise_exception:
                raise
            return Portal._get_access_status(e)
        if self._ignore_properties and items:
            delete_properties_from_dictionaries(items, self._ignore_properties)
        return items

    def access(self, query: str, metadata: bool = False, raw: bool = False, inserts: bool = False,
               report: bool = False, printf: Optional[Callable] = None) -> Portal.Access:
        access = Portal.Access.OK
        try:
            self.GET(query, metadata=metadata, raw=raw or inserts, raise_exception=True)
        except Exception as e:
            access = Portal._get_access_status(e)
        if report is True:
            Portal.report_access_status(access, query=query, verbose=False)
        return access

    @staticmethod
    def _get_access_status(e: Optional[Exception]) -> Access:
        # TODO
        # Bad status code for GET request for http://localhost:8000/a7ca8dab-b1f0-4f61-a892-87fe3516830a:
        # 403. Reason: {'@type': ['HTTPForbidden', 'Error'], 'status': 'error',
        # 'code': 403, 'title': 'Forbidden', 'description': 'Access was denied to this resource.',
        # 'detail': 'Unauthorized: item_view failed permission check'}
        # VS.
        # Bad status code for GET request for http://localhost:8000/a7ca8dab-b1f0-4f61-a892-87fe3516830aasfd:
        # 404. Reason: {\'@type\': [\'HTTPNotFound\', \'Error\'], \'status\': \'error\',
        # \'code\': 404, \'title\': \'Not Found\', \'description\':
        # \'The resource could not be found.\', \'detail\':
        # "debug_notfound of url http://localhost:8000/a7ca8dab-b1f0-4f61-a892-87fe3516830aasfd; path_info:
        # \'/a7ca8dab-b1f0-4f61-a892-87fe3516830aasfd\', context: <encoded.root.SMAHTRoot object at 0x1371d39b0>,
        # view_name: \'a7ca8dab-b1f0-4f61-a892-87fe3516830aasfd\', subpath: (), traversed: (),
        # root: <encoded.root.SMAHTRoot object at 0x1371d39b0>,
        # vroot: <encoded.root.SMAHTRoot object at 0x1371d39b0>, vroot_path: ()"}'
        def contains_word(text: str, word: str) -> bool:  # noqa
            return bool(re.search(rf"\b{word}\b", text, re.IGNORECASE))
        if isinstance(e, Exception):
            if (contains_word(e := str(e), "HTTPForbidden") or
                contains_word(e, "forbidden") or contains_word(e, "unauthorized")):  # noqa
                access = Portal.Access.NO_ACCESS
            elif contains_word(e, "HTTPNotFound") or (contains_word(e, "found") and
                                                      contains_word(e, "not") and contains_word(e, "404")):
                access = Portal.Access.NOT_FOUND
            else:
                access = Portal.Access.ERROR
        else:
            access = Portal.Access.OK
        return access

    @staticmethod
    def report_access_status(access: Portal.Access,
                             query: Optional[str] = None, verbose: bool = False,
                             printf: Optional[Callable] = None) -> None:
        if not callable(printf):
            printf = print
        if access == Portal.Access.NOT_FOUND:
            if verbose is True:
                printf(f"Portal items(s) not found{f': {query}' if query else '.'}")
            else:
                printf(f"{f'{query}: ' if query else ''}Not found")
        elif access == Portal.Access.NO_ACCESS:
            if verbose is True:
                printf(f"Forbidden access to Portal item(s){f': {query}' if query else '.'}")
            else:
                printf(f"{f'{query}: ' if query else ''}Forbidden")
        elif access == Portal.Access.ERROR:
            if verbose is True:
                printf(f"Error retrieving Portal item(s){f': {query}' if query else ''}")
            else:
                printf(f"{f'{query}: ' if query else ''}Error")
        elif access == Portal.Access.OK:
            printf(f"{query}: OK")
        else:
            if verbose is True:
                printf(f"Unknown error retrieving Portal item(s){f': {query}' if query else ''}")
            else:
                printf(f"{f'{query}: ' if query else ''}Unknown error")


def main() -> int:

    started = time.time()

    argv = ARGV({
        ARGV.OPTIONAL(str): ["query"],
        ARGV.OPTIONAL(bool): ["--ping"],
        ARGV.OPTIONAL(bool): ["--version"],
        ARGV.AT_LEAST_ONE_OF: ["query", "--ping", "--version"],
        ARGV.OPTIONAL(str): ["--app"],
        ARGV.OPTIONAL(str): ["--env", "--e"],
        ARGV.OPTIONAL(str): ["--ini", "--ini-file"],
        ARGV.OPTIONAL(bool): ["--inserts", "--insert"],
        ARGV.OPTIONAL(str): ["--output", "--out", "--o"],
        ARGV.OPTIONAL(bool): ["--check-access-only", "--check-access", "--check", "--access"],
        ARGV.OPTIONAL(bool): ["--raw"],
        ARGV.OPTIONAL(bool): ["--database"],
        ARGV.OPTIONAL(bool): ["--metadata"],
        ARGV.OPTIONAL(bool): ["--nometadata"],
        ARGV.OPTIONAL(bool): ["--noformat"],
        ARGV.OPTIONAL(bool): ["--deleted"],
        ARGV.OPTIONAL(bool): ["--json"],
        ARGV.OPTIONAL(bool): ["--yaml", "--yml"],
        ARGV.OPTIONAL(bool): ["--refs", "--ref"],
        ARGV.OPTIONAL(bool): ["--noignore-properties", "--noignore", "--no-ignore-properties", "--no-ignore" "--all"],
        ARGV.OPTIONAL([str]): ["--ignore-properties", "--ignore"],
        ARGV.OPTIONAL(bool): ["--nocruft"],
        ARGV.OPTIONAL(bool): ["--sort"],
        ARGV.OPTIONAL(bool): ["--reorganize", "--reorg"],
        ARGV.OPTIONAL(bool): ["--uuids"],
        ARGV.OPTIONAL([str]): ["--pick"],
        ARGV.OPTIONAL(str): ["--pick-separator", "--pick-sep"],
        ARGV.OPTIONAL(int): ["--limit", "--count"],
        ARGV.OPTIONAL(int): ["--offset", "--skip", "--from"],
        ARGV.OPTIONAL(bool): ["--append"],
        ARGV.OPTIONAL(bool): ["--merge"],
        ARGV.OPTIONAL(bool): ["--overwrite"],
        ARGV.OPTIONAL(bool): ["--randomize-md5sum-values"],
        ARGV.OPTIONAL(bool): ["--noscrub-sids"],
        ARGV.OPTIONAL(bool): ["--show"],
        ARGV.OPTIONAL(bool): ["--nowarnings", "--nowarning", "--nowarn"],
        ARGV.OPTIONAL(bool): ["--verbose"],
        ARGV.OPTIONAL(bool): ["--debug"],
        ARGV.OPTIONAL(int, 50): ["--nthreads", "--threads"],
        ARGV.OPTIONAL(bool): ["--sanity-check", "--sanity"],
        ARGV.OPTIONAL(bool): ["--timing", "--time", "--times"],
        ARGV.OPTIONAL(bool): ["--noheader"],
        ARGV.OPTIONAL(bool): ["--argv"],
        ARGV.AT_MOST_ONE_OF: ["--inserts", "--raw"],
        ARGV.AT_MOST_ONE_OF: ["--inserts-files", "--raw"],
        ARGV.AT_MOST_ONE_OF: ["--metadata", "--nometadata"],
        ARGV.OPTIONAL(bool): ["--exceptions", "--exception", "--except"],
        ARGV.DEPENDENCY: ["--no-ignore-properties", ARGV.DEPENDS_ON, ["--raw", "--inserts"]],
    })

    _setup_debugging(argv)

    if argv.version:
        print(f"hms-portal-read: {get_version()}")

    if not (portal := Portal.create(argv.env or argv.ini, app=argv.app, show=argv.show,
                                    verbose=argv.verbose and not argv.noheader, debug=argv.debug,
                                    ping=argv.ping, raise_exception=argv.exceptions, printf=_info)):
        return 1

    if not argv.query:
        return 0

    if ("/" not in argv.query) and ("-" not in argv.query):
        if argv.query.lower() == "schemas":
            _write_data(portal.get_schemas(), argv)
        elif schema := portal.get_schema(argv.query):
            _write_data(schema, argv)
            return 0

    if argv.noignore_properties:
        portal.ignore_properties = []
    elif argv.inserts:
        portal.ignore_properties = _ITEM_IGNORE_PROPERTIES_INSERTS

    if argv.nocruft:
        portal.ignore_properties = _ITEM_IGNORE_PROPERTIES_NOCRUFT
    if argv.ignore_properties:
        portal.ignore_properties = portal.ignore_properties + argv.ignore_properties

    # By default use Portal.get_metadata, iff the given query argument does not start with a slash,
    # otherwise use Portal.get; override to use portal.get_metadata with the --metadata
    # option or override to use portal.get with the --nometadata option.
    metadata = (argv.metadata or (not argv.query.startswith("/"))) and (not argv.nometadata)

    if (not metadata) and (not argv.query.startswith("/")):
        argv.query = f"/{argv.query}"

    _verbose(f"Querying Portal for item(s): {argv.query}")

    if argv.check_access_only:
        return 0 if portal.access(argv.query, metadata=metadata, raw=argv.raw, inserts=argv.inserts,
                                  report=True, printf=_print) == Portal.Access.OK else 1

    items = _portal_get(portal, argv.query, metadata=metadata, raw=argv.raw, inserts=argv.inserts,
                        limit=argv.limit, offset=argv.offset, database=argv.database,
                        deleted=argv.deleted, nthreads=argv.nthreads)
    if isinstance(items, Portal.Access):
        portal.report_access_status(items, query=argv.query, verbose=True)
        return 1

    if graph := items.get("@graph"):
        items = graph
    if isinstance(items, dict):
        items = [items]

    if argv.refs:
        items.extend(_get_portal_referenced_items(
            portal, items, raw=argv.raw, metadata=argv.metadata,
            inserts=argv.inserts, database=argv.database, nthreads=argv.nthreads))

    if argv.inserts:
        items = _insertize_items(items)

    if argv.sort:
        items = sort_dictionary(items)

    if argv.reorganize:
        # Simply put uuid and @type at the top of each item.
        def reorganize_item(item: dict) -> None:  # noqa
            if isinstance(item, dict):
                if (item_type := item.get("@type")) is not None:
                    del item["@type"]
                    item = {"@type": item_type, **item}
                if (item_uuid := item.get(_ITEM_UUID_PROPERTY_NAME)) is not None:
                    del item[_ITEM_UUID_PROPERTY_NAME]
                    item = {_ITEM_UUID_PROPERTY_NAME: item_uuid, **item}
            return item
        def reorganize_items(items: List[dict]) -> None:  # noqa
            if isinstance(items, list):
                for index in range(len(items)):
                    items[index] = reorganize_item(items[index])
        if isinstance(items, list):
            reorganize_items(items)
        elif isinstance(items, dict):
            for item_type in items:
                reorganize_items(items[item_type])

    if not argv.noscrub_sids:
        # If for some reasone there are "sid" values in the retrieved items then we remove them, by default.
        _scrub_sids_from_items(items)

    if argv.randomize_md5sum_values:
        _randomize_md5sum_values(items)

    if argv.uuids or argv.pick:
        if argv.pick:
            if _ITEM_UUID_PROPERTY_NAME not in argv.pick:
                argv.pick.insert(0, _ITEM_UUID_PROPERTY_NAME)
        else:
            argv.pick = [_ITEM_UUID_PROPERTY_NAME]
        if argv.pick_separator:
            if argv.pick_separator in ["\\t", "tab", "TAB"]:
                argv.pick_separator = "\t"
        def write_item_properties(items: List[dit], names: List[str],  # noqa
                                  write: Callable, separator: Optional[str] = None) -> None:
            if isinstance(items, list) and isinstance(names, list) and callable(write):
                if not isinstance(separator, str):
                    separator = " "
                for item in items:
                    values = []
                    for name in names:
                        if value := get_property(item, name):
                            values.append(str(value))
                    write(separator.join(values))
        if not isinstance(items, list):
            _error("Result set if not a list of items which is required when using the --uuids option.")
        if argv.output:
            if os.path.isdir(argv.output):
                _error(f"Specified output file already exists as a directory: {argv.output}")
            elif os.path.exists(argv.output):
                if (not argv.append) and (argv.overwrite or yes_or_no("Overwrite this file?")):
                    with io.open(argv.output, "w") as f:
                        write_item_properties(items, argv.pick,
                                              write=lambda value: f.write(f"{value}\n"), separator=argv.pick_separator)
                elif argv.append or yes_or_no("Append to this file?"):
                    with io.open(argv.output, "a") as f:
                        write_item_properties(items, argv.pick,
                                              write=lambda value: f.write(f"{value}\n"), separator=argv.pick_separator)
                else:
                    return 1
        else:
            write_item_properties(items, argv.pick, write=print, separator=argv.pick_separator)
        return 0

    elif argv.output:
        if argv.inserts and (os.path.isdir(argv.output) or argv.output.endswith(os.sep)):
            _write_inserts_output_files(items, argv.output, noformat=argv.noformat,
                                        overwrite=argv.overwrite, merge=argv.merge, append=argv.append)
        else:
            _write_output_file(items, argv.output, inserts=argv.inserts, noformat=argv.noformat,
                               overwrite=argv.overwrite, merge=argv.merge, append=argv.append)

    else:
        _write_data(items, argv)

    if argv.refs and argv.sanity_check:
        _verbose("Sanity checking for missing referenced items ...", end="")
        started_sanity_check = time.time()
        missing_items_referenced = _sanity_check_missing_items_referenced(items)
        duration_sanity_check = format_duration(time.time() - started_sanity_check)
        if missing_items_referenced:
            _verbose(f" {duration_sanity_check}" if duration_sanity_check != "00:0000" else "")
            _error(f"Missing items for referenced items found: {len(missing_items_referenced)}", exit=False)
            if argv.verbose:
                _verbose(missing_items_referenced)
        else:
            _verbose(f" OK {chars.check}{f' {duration_sanity_check}' if duration_sanity_check != '00:00:00' else ''}")

    if argv.verbose:
        uuids_count = len(get_uuids(items))
        referenced_uuids_count = len(get_referenced_uuids(items, exclude_properties=_ITEM_IGNORE_REF_PROPERTIES))
        _verbose(f"Total items fetched:"
                 f" {uuids_count}"
                 f"{f' {chars.dot} refs: {referenced_uuids_count}' if referenced_uuids_count != uuids_count else ''}")
        if argv.verbose and argv.inserts and isinstance(items, dict):
            type_count = len(set(items.keys()))
            _verbose(f"Total item types fetched: {type_count}")
    if argv.timing or argv.debug:
        _info(f"Calls to portal.get_metadata: {portal.get_metadata_call_count}"
              f" {chars.dot} {format_duration(portal.get_metadata_call_duration)}")
        _info(f"Calls to portal.get: {portal.get_call_count} {chars.dot} {format_duration(portal.get_call_duration)}")
    if argv.verbose or argv.timing or argv.debug:
        duration = time.time() - started
        _info(f"Duration: {format_duration(duration)}")
    return 0


def _write_inserts_output_files(items: dict, output_directory: str, noformat: bool = False,
                                overwrite: bool = False, merge: bool = False, append: bool = False) -> None:
    os.makedirs(output_directory, exist_ok=True)
    for item_type in items:
        item_type_items = items[item_type]
        output_file = os.path.join(output_directory, f"{to_snake_case(item_type)}.json")
        overwrite_output_file = False
        merge_into_output_file = False
        append_to_output_file = False
        if os.path.exists(output_file):
            if overwrite:
                overwrite_output_file = True
            elif merge:
                merge_into_output_file = True
            elif append:
                append_to_output_file = True
            else:
                _print(f"Specified output file already exists: {output_file} {chars.dot}")
                if not yes_or_no("Merge into this file?"):
                    if not yes_or_no("Overwrite this file?"):
                        if not yes_or_no("Append to this file?"):
                            continue
                        append_to_output_file = True
                    else:
                        overwrite_output_file = True
                else:
                    merge_into_output_file = True
            if merge_into_output_file or append_to_output_file:
                try:
                    with io.open(output_file, "r") as f:
                        if not isinstance(existing_items := json.load(f), list):
                            _warning(f"JSON file does not contain a list: {output_file}")
                            continue
                except Exception:
                    _warning(f"Cannot load file as JSON: {output_file}")
                    continue
                if merge_into_output_file:
                    merge_identifying_property_name = _ITEM_UUID_PROPERTY_NAME
                    for item in item_type_items:
                        index = find_dictionary_item(existing_items,
                                                     property_value=item.get(merge_identifying_property_name),
                                                     property_name=merge_identifying_property_name)
                        if index is not None:
                            existing_items[index] = item
                        else:
                            existing_items.append(item)
                elif append_to_output_file:
                    existing_items.extend(item_type_items)
                item_type_items = existing_items
        with io.open(output_file, "w") as f:
            if overwrite_output_file:
                _debug(f"Overwriting output file: {output_file}")
            elif merge_into_output_file:
                _debug(f"Merging into output file: {output_file}")
            else:
                _debug(f"Writing output file: {output_file}")
            json.dump(item_type_items, f, indent=None if (noformat is True) else 4)
            if overwrite_output_file:
                _verbose(f"Output file overwritten: {output_file}")
            elif merge_into_output_file:
                _verbose(f"Output file merged into: {output_file}")
            elif append_to_output_file:
                _verbose(f"Output file appended to: {output_file}")
            else:
                _verbose(f"Output file written: {output_file}")


def _write_output_file(items: dict, output_file: str, inserts: bool = False, noformat: bool = False,
                       overwrite: bool = False, merge: bool = False, append: bool = False) -> None:
    overwrite_output_file = False
    merge_into_output_file = False
    append_to_output_file = False
    if os.path.isdir(output_file):
        _error(f"Specified output file already exists as a directory: {output_file}")
    if os.path.exists(output_file):
        if overwrite:
            overwrite_output_file = True
        elif merge:
            merge_into_output_file = True
        elif append:
            append_to_output_file = True
        else:
            _print(f"Specified output file already exists: {output_file}")
            if not yes_or_no("Merge into this file?"):
                if not yes_or_no("Overwrite this file?"):
                    if not yes_or_no("Append to this file?"):
                        return 1
                    append_to_output_file = True
                else:
                    overwrite_output_file = True
            else:
                merge_into_output_file = True
        if merge_into_output_file or append_to_output_file:
            try:
                with io.open(output_file, "r") as f:
                    existing_items = json.load(f)
                    if inserts:
                        if not isinstance(existing_items, dict):
                            _error(f"JSON file does not contain a dictionary: {output_file}")
                    elif not isinstance(existing_items, list):
                        _error(f"JSON file does not contain a list: {output_file}")
            except Exception:
                _error(f"Cannot load file as JSON: {output_file}")
            if merge_into_output_file:
                merge_identifying_property_name = _ITEM_UUID_PROPERTY_NAME
                if inserts:
                    for item_type in items:
                        if existing_item_type_items := existing_items.get(item_type):
                            if isinstance(existing_item_type_items, list):
                                for item in items[item_type]:
                                    index = find_dictionary_item(
                                        existing_item_type_items,
                                        property_value=item.get(merge_identifying_property_name),
                                        property_name=merge_identifying_property_name)
                                    if index is not None:
                                        existing_item_type_items[index] = item
                                    else:
                                        existing_item_type_items.append(item)
                        else:
                            existing_items[item_type] = items[item_type]
                else:
                    for item in items:
                        index = find_dictionary_item(
                            existing_items,
                            property_value=item.get(merge_identifying_property_name),
                            property_name=merge_identifying_property_name)
                        if index is not None:
                            existing_items[index] = item
                        else:
                            existing_items.append(item)
            elif append_to_output_file:
                if inserts:
                    for item_type in items:
                        if existing_item_type_items := existing_items.get(item_type):
                            if isinstance(existing_item_type_items, list):
                                existing_item_type_items.append(items[item_type])
                        else:
                            existing_items[item_type] = items[item_type]
                else:
                    existing_items.extend(items)
            items = existing_items
    with io.open(output_file, "w") as f:
        json.dump(items, f, indent=None if noformat else 4)
    if overwrite_output_file:
        _verbose(f"Output file overwritten: {output_file}")
    elif merge_into_output_file:
        _verbose(f"Output file merged into: {output_file}")
    elif append_to_output_file:
        _verbose(f"Output file appended to: {output_file}")
    else:
        _verbose(f"Output file written: {output_file}")


def _write_data(data: Any, argv: ARGV) -> None:
    if argv.yaml is True:
        _print(yaml.dump(data).strip())
    elif argv.noformat is True:
        _print(data)
    else:
        _print(json.dumps(data, indent=4))


def _get_portal_referenced_items(portal: Portal, item: dict, metadata: bool = False,
                                 raw: bool = False, inserts: bool = False, database: bool = False,
                                 nthreads: Optional[int] = None) -> List[dict]:
    referenced_items = [] ; ignore_uuids = [] ; referenced_items_batch = item ; referenced_uuids_last = None  # noqa
    if isinstance(item, dict):
        debug_item = item.get(_ITEM_UUID_PROPERTY_NAME)
    elif isinstance(item, list):
        if (len(item) == 1) and isinstance(item[0], dict):
            debug_item = item[0].get(_ITEM_UUID_PROPERTY_NAME)
        else:
            debug_item = f"{len(item)} items"
    else:
        debug_item = "unknown"
    _debug(f"Retrieving items referenced by:"
           f" {debug_item} {chars.dot} ignored: {len(ignore_uuids)}")
    while referenced_uuids := get_referenced_uuids(referenced_items_batch, ignore_uuids=ignore_uuids,
                                                   exclude_uuid=True, include_paths=True,
                                                   exclude_properties=_ITEM_IGNORE_REF_PROPERTIES):
        _debug(f"Retrieved UUIDs referenced by {debug_item}:"
               f" {len(referenced_uuids)} {chars.dot} ignored: {len(ignore_uuids)}")
        if (referenced_uuids := set(referenced_uuids)) == referenced_uuids_last:
            _debug(f"No more refrenced UUID: {len(referenced_uuids)}"
                   f"{chars.dot} ignored: {len(ignore_uuids)} {chars.dot} {len(referenced_uuids)}")
            break
        referenced_uuids_last = referenced_uuids
        referenced_items_batch = _get_portal_items_for_uuids(
            portal, referenced_uuids, metadata=metadata, raw=raw, inserts=inserts, database=database, nthreads=nthreads)
        for referenced_item in referenced_items_batch:
            if isinstance(referenced_item, dict):
                referenced_items.append(referenced_item)
        for item in referenced_items_batch:
            if isinstance(item, dict):
                if item_uuid := item.get(_ITEM_UUID_PROPERTY_NAME):
                    if item_uuid not in ignore_uuids:
                        ignore_uuids.append(item_uuid)
    _debug(f"Retrieved items referenced by {debug_item}: {len(referenced_items)}")
    return referenced_items


def _get_portal_items_for_uuids(portal: Portal, uuids: Union[List[str], Set[str], Tuple[str], str],
                                metadata: bool = False, raw: bool = False, inserts: bool = False,
                                database: bool = False, nthreads: Optional[int] = None) -> List[dict]:
    items = []
    if not isinstance(uuids, (list, set, tuple)):
        if not is_uuid(uuids):
            return []
        uuids = [uuids]
    fetch_portal_item_functions = []
    def fetch_portal_item(uuid: str) -> Optional[dict]:  # noqa
        nonlocal portal, metadata, raw, database, items
        if item := _portal_get(portal, uuid, metadata=True,
                               raw=raw, inserts=inserts, database=database, nthreads=nthreads):
            if not isinstance(item, dict):
                _debug(f"Cannot retrieve item: {uuid} [{item}]")
            items.append(item)  # TODO: make thread-safe.
    if fetch_portal_item_functions := [lambda uuid=uuid: fetch_portal_item(uuid) for uuid in uuids if is_uuid(uuid)]:
        run_concurrently(fetch_portal_item_functions, nthreads=nthreads)
    return items


@lru_cache(maxsize=1024)
def _portal_get(portal: Portal, query: str, metadata: bool = False, raw: bool = False,
                inserts: bool = False, database: bool = False,
                limit: Optional[int] = None, offset: Optional[int] = None, deleted: bool = False,
                nthreads: Optional[int] = None) -> dict:
    if inserts is True:
        return _portal_get_inserts(portal, query, metadata=metadata, database=database,
                                   limit=limit, offset=offset, deleted=deleted, nthreads=nthreads)
    return portal.GET(query, metadata=metadata, raw=raw,
                      database=database, limit=limit, offset=offset, deleted=deleted)


@lru_cache(maxsize=1024)
def _portal_get_inserts(portal: Portal, query: str, metadata: bool = False, database: bool = False,
                        limit: Optional[int] = None, offset: Optional[int] = None,
                        deleted: bool = False,
                        nthreads: Optional[int] = None) -> dict:
    item = None ; item_noraw = None  # noqa
    def fetch_portal_item() -> None:  # noqa
        nonlocal portal, query, database, item
        item = portal.GET(query, metadata=metadata, raw=True,
                          database=database, limit=limit, offset=offset, deleted=deleted)
    def fetch_portal_item_noraw() -> None:  # noqa
        # This is to get the non-raw frame item format which has the type information (the raw frame does not).
        nonlocal portal, query, metadata, database, item_noraw
        item_noraw = portal.GET(query, metadata=metadata, raw=False,
                                database=database, limit=limit, offset=offset, deleted=deleted,
                                field=_ITEM_UUID_PROPERTY_NAME)
    run_concurrently([fetch_portal_item, fetch_portal_item_noraw], nthreads=min(2, nthreads))
    if not item:
        return {}
    elif item in [Portal.Access.NOT_FOUND, Portal.Access.NO_ACCESS, Portal.Access.ERROR]:
        return item
    if isinstance(graph := item.get("@graph"), list):
        if isinstance(item_noraw_graph := item_noraw.get("@graph"), list):
            def find_item_type(uuid: str) -> Optional[str]:
                nonlocal item_noraw, item_noraw_graph
                for item_element in item_noraw_graph:
                    if item_element.get(_ITEM_UUID_PROPERTY_NAME) == uuid:
                        return Portal.get_item_type(item_element)
                return None
            for graph_item in graph:
                if graph_item_type := find_item_type(graph_item.get(_ITEM_UUID_PROPERTY_NAME)):
                    graph_item[_ITEM_TYPE_PSEUDO_PROPERTY_NAME] = graph_item_type
    else:
        if item_type := Portal.get_item_type(item_noraw):
            item[_ITEM_TYPE_PSEUDO_PROPERTY_NAME] = item_type
    return item


def _insertize_items(items: dict) -> dict:
    items_by_type = {}
    for item in items:
        if isinstance(item, dict):
            item_type = item.get(_ITEM_TYPE_PSEUDO_PROPERTY_NAME)
            if not (item_list := items_by_type.get(item_type)):
                items_by_type[item_type] = (item_list := [])
            item_list.append(item)
            item.pop(_ITEM_TYPE_PSEUDO_PROPERTY_NAME, None)
    return items_by_type


def _randomize_md5sum_values(items: Union[List[dict], dict]) -> None:
    global _ITEM_MD5SUM_PROPERTY_NAME
    def create_random_md5() -> str:  # noqa
        return hashlib.md5(str(uuid4()).encode()).hexdigest()
    if isinstance(items, list):
        for element in items:
            _randomize_md5sum_values(element)
    elif isinstance(items, dict):
        for key in items:
            if key == _ITEM_MD5SUM_PROPERTY_NAME:
                items[key] = create_random_md5()
            _randomize_md5sum_values(items[key])


def _scrub_sids_from_items(items: dict) -> None:
    if isinstance(items, list):
        for element in items:
            _scrub_sids_from_items(element)
    elif isinstance(items, dict):
        for key in list(items.keys()):
            if key == _ITEM_SID_PROPERTY_NAME:
                _warning(f"Deleting sid from item:"
                         f" {uuid if (uuid := items.get(_ITEM_UUID_PROPERTY_NAME)) else 'unknown'}")
                del items[key]
            else:
                _scrub_sids_from_items(items[key])


def _sanity_check_missing_items_referenced(data: Union[dict, list]) -> List[str]:
    missing_items_referenced = []
    for referenced_uuid in get_referenced_uuids(data, exclude_properties=_ITEM_IGNORE_REF_PROPERTIES):
        if not contains_uuid(data, referenced_uuid):
            missing_items_referenced.append(referenced_uuid)
    return missing_items_referenced


def _print(*args, **kwargs) -> None:
    print(*args, **kwargs)


def _info(*args, **kwargs) -> None:
    _print(*args, **kwargs, file=sys.stderr, flush=True)


def _verbose(*args, **kwargs) -> None:
    _print(*args, **kwargs, file=sys.stderr, flush=True)


def _debug(message: str) -> None:
    _print(f"DEBUG: {message}", file=sys.stderr, flush=True)


def _warning(message: str) -> None:
    _print(f"WARNING: {message}", file=sys.stderr, flush=True)


def _error(message: str, exit: bool = True) -> None:
    print(f"ERROR: {message}", file=sys.stderr, flush=True)
    if exit is not False:
        sys.exit(1)


def _nofunction(*args, **kwargs) -> None:
    pass


def _setup_debugging(argv: ARGV) -> None:

    global _verbose, _debug, _nofunction
    if argv.nowarnings: _warning = _nofunction  # noqa
    if not argv.verbose: _verbose = _nofunction  # noqa
    if not argv.debug: _debug = _nofunction  # noqa

    original_requests_get = requests.get
    def requests_get(*args, **kwargs):  # noqa
        if isinstance(args, tuple) and (len(args) > 0):
            message = f"{args[0]}"
        else:
            message = str(args)
        if ((kwargs.get("headers", {}).get("Content-type") == "application/json") and
            (kwargs.get("headers", {}).get("Accept") == "application/json") and
            isinstance(auth := kwargs.get("auth"), tuple) and (len(auth) > 0) and (auth := auth[0])):  # noqa
            message += f" {chars.dot} standard headers {chars.dot} key: {auth[0:2] + '******'}"
        else:
            message += f" {chars.dot} {str(kwargs)}"
        _debug(f"request.get: {message}")
        return original_requests_get(*args, **kwargs)
    requests.get = requests_get

    if argv.argv:
        _print(json.dumps(argv._dict, indent=4))


if __name__ == "__main__":
    status = main()
    sys.exit(status if isinstance(status, int) else 0)



import iterm2
import datetime


# This function gets called once per second.
@iterm2.StatusBarRPC
async def coro(knobs):
    return datetime.datetime.now(datetime.timezone.utc).strftime("UTC:%H:%M:%S")


def iterm_run() -> None:
    async def main(connection):
        component = iterm2.StatusBarComponent(
            short_description="UTC Clock",
            detailed_description="Shows the time in jolly old England",
            knobs=[],
            exemplar="UTC: 12:00",
            update_cadence=1,
            identifier="com.iterm2.example.utc-clock")
        await component.async_register(connection, coro)
    iterm2.run_forever(main)


iterm_run()



from hms_utils.portal.portal_utils import Portal
from hms_utils.argv import ARGV


def main():

    argv = ARGV({
        ARGV.REQUIRED(str): "--env",
        ARGV.REQUIRED([str]): "uuids"
    })

    portal = Portal(argv.env)
    if portal.reindex_metadata(argv.uuids):
        print("OK")
    else:
        print("ERROR")


if __name__ == "__main__":
    main()
    pass



import hashlib
import io  # noqa
import json  # noqa
import os
import sys
from typing import List, Union
from uuid import uuid4
from dcicutils.command_utils import yes_or_no


def main():
    if len(sys.argv) > 1:
        file = sys.argv[1]
        randomize_md5sum_values_in_json_file(file, verbose=True)


def randomize_md5sum_values(items: Union[List[dict], dict]) -> None:
    def create_random_md5() -> str:
        return hashlib.md5(str(uuid4()).encode()).hexdigest()
    if isinstance(items, list):
        for element in items:
            randomize_md5sum_values(element)
    elif isinstance(items, dict):
        for key in items:
            if key == "md5sum":
                items[key] = create_random_md5()
            randomize_md5sum_values(items[key])


def randomize_md5sum_values_in_json_file(file: str, overwrite: bool = False, verbose: bool = False) -> bool:
    if not os.path.exists(file):
        if verbose:
            print(f"File does not exist: {file}", file=sys.stderr, flush=True)
        return False
    try:
        with io.open(file, "r") as f:
            try:
                if not (data := json.load(f)):
                    print(f"Cannot load JSON from file: {file}", file=sys.stderr, flush=True)
                    return False
                randomize_md5sum_values(data)
            except Exception:
                print(f"Exception loading JSON from file: {file}", file=sys.stderr, flush=True)
                return False
        if overwrite is not True:
            if verbose is not True:
                return False
            if not yes_or_no(f"Overwrite file: {file} ?"):
                return False
        if verbose is True:
            print(f"Overwriting file: {file}")
        try:
            with io.open(file, "w") as f:
                json.dump(data, f, indent=4)
        except Exception:
            print(f"Exception writing file: {file}")
            return False
    except Exception:
        print(f"Cannot open file: {file}")
        return False


if __name__ == "__main__":
    status = main()
    sys.exit(status if isinstance(status, int) else 0)



from dcicutils.ff_utils import search_metadata
from hms_utils.portal.portal_utils import Portal
from hms_utils.argv import ARGV
from hms_utils.chars import chars


def main():

    argv = ARGV({
        ARGV.REQUIRED(str): "--env",
        ARGV.OPTIONAL(bool): ["--details", "--detail", "--verbose"],
        ARGV.OPTIONAL(bool): "--sid",
        ARGV.OPTIONAL(bool): "--sid-reindex",
        ARGV.OPTIONAL([str]): "uuids"
    })

    portal = Portal(argv.env)
    status = 0
    uuids_with_sid_error = []

    if argv.uuids:
        for uuid in argv.uuids:
            if item := portal.get_metadata(uuid):
                if _check_for_validation_errors(item, sid=argv.sid):
                    if argv.sid and argv.sid_reindex:
                        uuids_with_sid_error.append(item.get("uuid"))
                    status

    else:
        query = "/search/?type=Item&validation_errors.name%21=No+value&limit=100000"
        ignored_types = []
        for item in search_metadata(query, key=portal.key, is_generator=True):
            if isinstance(item, dict) and (item_uuid := item.get("uuid")):
                if (item_type := portal.get_schema_type(item)) not in ignored_types:
                    print(f"{item_uuid}: {item_type}")
                    status = 1
                    if argv.sid or argv.sid_reindex or argv.details:
                        if item := portal.get_metadata(item_uuid):
                            if _check_for_validation_errors(item, sid=argv.sid):
                                if argv.sid and argv.sid_reindex:
                                    uuids_with_sid_error.append(item.get("uuid"))

    if uuids_with_sid_error:
        for uuid_with_sid_error in uuids_with_sid_error:
            print(f"Reindexing item with sid error : {uuid_with_sid_error}")
            portal.reindex_metadata(uuid_with_sid_error)

    exit(status)


def _check_for_validation_errors(item: dict, sid: bool = False) -> bool:
    found = False
    if errors := item.get("validation-errors"):
        for error in errors:
            if (sid is not True) or ("sid" in str(error)):
                found = True
                print(f"{chars.rarrow_hollow} ERROR: {error.get('description')}")
    return found


if __name__ == "__main__":
    main()
    pass



# Script to get all possible URLs from all AWS accounts.
# TODO: Does not even include for example: wolf.smaht.org

import boto3
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import socket
import warnings
from hms_utils.aws_env import AwsProfiles

warnings.simplefilter('ignore', InsecureRequestWarning)


def _print(s):
    print(s, flush=True)


def list_elb_dns(account_id):
    session = boto3.Session(profile_name=account_id)
    elb_client = session.client('elbv2')
    response = elb_client.describe_load_balancers()
    elb_urls = []
    for lb in response['LoadBalancers']:
        dns = lb['DNSName']
        elb_urls.append(dns)
    return sorted(elb_urls)


def list_api_gateway_urls(account_id):
    session = boto3.Session(profile_name=account_id)
    api_gateway_client = session.client('apigateway')
    apis = api_gateway_client.get_rest_apis(limit=1000)
    api_urls = []
    for api in apis['items']:
        api_urls.append(f"{api['id']}.execute-api.{session.region_name}.amazonaws.com")
    return sorted(api_urls)


def list_cloudfront_urls(account_id):
    session = boto3.Session(profile_name=account_id)
    cloudfront_client = session.client('cloudfront')
    response = cloudfront_client.list_distributions()
    cf_urls = []
    if response := response['DistributionList']:
        for dist in response.get('Items', []):
            cf_urls.append(dist['DomainName'])
    return sorted(cf_urls)


def list_route53_urls(account_id):
    session = boto3.Session(profile_name=account_id)
    route53_client = session.client('route53')
    zones = route53_client.list_hosted_zones()
    all_urls = []
    for zone in zones['HostedZones']:
        records = route53_client.list_resource_record_sets(HostedZoneId=zone['Id'])
        for record in records['ResourceRecordSets']:
            if record['Type'] in ['A', 'CNAME']:
                if (dns := record['Name']).endswith("."):
                    dns = dns[:-1]
                all_urls.append(dns)
    return sorted(all_urls)


def list_acm_certificates(account_id):
    session = boto3.Session(profile_name=account_id)
    client = session.client('acm')
    paginator = client.get_paginator('list_certificates')
    urls = []
    for page in paginator.paginate():
        for cert in page['CertificateSummaryList']:
            cert_arn = cert['CertificateArn']
            response = client.describe_certificate(CertificateArn=cert_arn)
            certificate = response['Certificate']
            domains = certificate['DomainName']
            subject_alternative_names = certificate.get('SubjectAlternativeNames', [])
            all_domains = [domains] + subject_alternative_names
            urls += all_domains
    return list(set(urls))


def resolve_cname(url):
    try:
        cname = socket.gethostbyname_ex(url)
        return cname[0] if cname else None
    except socket.gaierror:
        return None


def ping(dns):
    http = False
    https = False
    skip = False
    forbidden = False
    if dns.startswith("\\052."):
        skip = True
        return http, https, skip, forbidden
    elif dns.endswith(".dev"):
        skip = True
        return http, https, skip, forbidden
    elif dns.endswith(".app"):
        skip = True
        return http, https, skip, forbidden
    try:
        response = requests.get(f"http://{dns}", allow_redirects=False, verify=False, timeout=4)
        if response.status_code == 403:
            forbidden = True
        http = True
    except Exception:
        pass
    try:
        response = requests.get(f"https://{dns}", verify=False, timeout=4)
        if response.status_code == 403:
            forbidden = True
        https = True
    except Exception:
        pass
    return http, https, skip, forbidden


def redirect_url(dns):
    if dns.startswith("\\052."):
        return None
    elif dns.endswith(".dev"):
        return None
    elif dns.endswith(".app"):
        return None
    response = None
    try:
        response = requests.get(f"http://{dns}", allow_redirects=False, verify=False, timeout=4)
    except Exception:
        pass
    try:
        requests.get(f"https://{dns}", verify=False, timeout=4)
    except Exception:
        return None
    redirect_url = None
    if response and (response.status_code in [301, 302, 303, 307, 308]):
        if redirect_url := response.headers['location']:
            if redirect_url.endswith("/"):
                redirect_url = redirect_url[:-1]
            if redirect_url.endswith(":443"):
                redirect_url = redirect_url[:-4]
            if redirect_url.startswith("http://"):
                redirect_url = redirect_url[7:]
            elif redirect_url.startswith("https://"):
                redirect_url = redirect_url[8:]
            if redirect_url != dns:
                return redirect_url
    return redirect_url


def ping_suffix(dns):
    http, https, skip, forbidden = ping(dns)
    if skip:
        result = " (N/A)"
    if http:
        if https:
            result = " (HTTP/HTTPS)"
        else:
            result = " (HTTP)"
    elif https:
        result = " (HTTPS)"
    else:
        result = " (UNREACHABLE)"
    if forbidden:
        result += " (FORBIDDEN)"
    return result


def format_url(dns):
    http, https, skip, forbidden = ping(dns)
    result = ""
    if skip:
        result = f"{dns} (N/A)"
    if http:
        if https:
            result = f"https://{dns} (HTTP/HTTPS)"
        else:
            result = f"http://{dns} (HTTP)"
    elif https:
        result = f"https://{dns} (HTTPS)"
    else:
        result = f"{dns} (UNREACHABLE)"
    if forbidden:
        result += " (FORBIDDEN)"
    return result


def just_url(s):
    if index := s.find(" "):
        return s[0:index]
    return s


aws_profiles = AwsProfiles.read()
for aws_profile in aws_profiles:
    _print(f"- ACCOUNT: {aws_profile.account} ({aws_profile.name})")
    if urls := list_acm_certificates(aws_profile.name):
        _print(f"  - CERTIFICATES: {len(urls)}")
        for url in urls:
            _print(f"    - {format_url(url)}")
            if (rurl := resolve_cname(url)) and (rurl != url):
                _print(f"      - CNAME: {format_url(rurl)}")
    if urls := list_elb_dns(aws_profile.name):
        _print(f"  - LOAD BALANCER: {len(urls)}")
        for url in urls:
            _print(f"    - {format_url(url)}")
            if (rurl := resolve_cname(url)) and (rurl != url):
                _print(f"      - CNAME: {format_url(rurl)}")
            if (rurl := redirect_url(url)) and (rurl != url):
                _print(f"      - REDIRECT: {format_url(rurl)}")
    if urls := list_api_gateway_urls(aws_profile.name):
        _print(f"  - API GATEWAY: {len(urls)}")
        for url in urls:
            _print(f"    - {format_url(url)}")
            if (rurl := resolve_cname(url)) and (rurl != url):
                _print(f"      - CNAME: {format_url(rurl)}")
    if urls := list_route53_urls(aws_profile.name):
        _print(f"  - ROUTE 53: {len(urls)}")
        for url in urls:
            _print(f"    - {format_url(url)}")
            if (rurl := resolve_cname(url)) and (rurl != url):
                _print(f"      - CNAME: {format_url(rurl)}")
    if urls := list_cloudfront_urls(aws_profile.name):
        _print(f"  - CLOUD FRONT: {len(urls)}")
        for url in urls:
            _print(f"    - {format_url(url)}")
            if (rurl := resolve_cname(url)) and (rurl != url):
                _print(f"      - CNAME: {format_url(rurl)}")



from datetime import date, datetime, timedelta
import json
from prettytable import PrettyTable, HRuleStyle as PrettyTableHorizontalStyle
import sys
from typing import List, Optional, Tuple, Union
from dcicutils.datetime_utils import format_date, format_time
from dcicutils.misc_utils import format_size
from hms_utils.argv import ARGV
from hms_utils.chars import chars
from hms_utils.datetime_utils import parse_datetime_string
from hms_utils.dictionary_utils import get_property
from hms_utils.portal.portal_utils import Portal
from hms_utils.type_utils import to_non_empty_string_list


def main():

    argv = ARGV({
        ARGV.OPTIONAL(str, "smaht-local"): ["--env"],
        ARGV.OPTIONAL(str): ["--app"],
        ARGV.OPTIONAL(str): ["--from-date", "--from"],
        ARGV.OPTIONAL(str): ["--thru-date", "--thru", "--to"],
        ARGV.OPTIONAL(bool): ["--all-dates", "--any-dates", "--any-date", "--all", "--any"],
        ARGV.OPTIONAL(int, 200): ["--limit", "--count"],
        ARGV.OPTIONAL(int, 0): ["--offset", "--skip"],
        ARGV.OPTIONAL(str, "released"): ["--status", "--state"],
        ARGV.OPTIONAL(bool, True): ["--noqc", "--noqcs"],
        ARGV.OPTIONAL(int, 1): ["--nmonths", "--months"],
        ARGV.OPTIONAL(bool, True): ["--group", "--grouped", "--group-month", "--grouped-month"],
        ARGV.OPTIONAL(str): ["--nosort"],
        ARGV.OPTIONAL(bool): ["--sort-reverse", "--sort-reversed", "--reverse", "--reversed"],
        ARGV.OPTIONAL(bool): ["--verbose"],
        ARGV.OPTIONAL(bool): ["--debug"],
        ARGV.OPTIONAL(bool): ["--dump"],
        ARGV.OPTIONAL(bool): ["--nowarnings", "--nowarning", "--nowarn"],
        ARGV.OPTIONAL(bool): "--ping"
    })

    _setup_debugging(argv)

    if not (portal := Portal.create(argv.env, app=argv.app, verbose=argv.verbose, debug=argv.debug, ping=argv.ping)):
        return 1

    from_date, thru_date = _parse_from_and_thru_date_args(argv) if not argv.all_dates else (None, None)

    date_property_name = "file_status_tracking.released"
    if status := argv.status:
        if (status := status.lower()) == "released":
            date_property_name = "file_status_tracking.released"
        elif status == "public":
            date_property_name = "file_status_tracking.public"
        elif status == "restricted":
            date_property_name = "file_status_tracking.restricted"
        elif status == "uploading":
            date_property_name = "file_status_tracking.uploading"
        elif status == "uploaded":
            date_property_name = "file_status_tracking.uploaded"
        else:
            date_property_name = "last_modified.date_modified"
            # _error(f"Invalid status specified: {argv.status}")

    # FYI: Note that using a portal query like this (using just date no time):
    #
    # - file_status_tracking.released.from=2024-11-01&file_status_tracking.released.to=2024-11-30
    #
    # results in an OpenSearch range query predicate like this:
    #
    # - {"embedded.file_status_tracking.released": {
    #    "format": "yyyy-MM-dd HH:mm", "gte": "2024-11-01 00:00", "lte": "2024-11-30 23:59"}}
    #
    # i.e. so adding 23:59 to mark the and of the day of the upper bound is properly handled by portal.

    query = "&".join(to_non_empty_string_list([
        f"/files",
        f"limit={argv.limit}",
        f"sort={'' if argv.sort_reverse else '-'}{date_property_name}" if not argv.nosort else None,
        f"status={status}" if status else None,
        f"data_category!=Quality Control" if argv.noqc else None,
        f"{date_property_name}.from={from_date}" if from_date else None,
        f"{date_property_name}.to={thru_date}" if thru_date else None,
        f"additional_facet=file_sets.libraries.analytes.samples.sample_sources.cell_line.code"
    ])).replace("&", "?", 1)

    # /search/?type=OutputFile&additional_facet=file_sets.libraries.analytes.samples.sample_sources.cell_line.code&status=released&file_status_tracking.released.from=2024-11-20&format=json

    if argv.debug:
        _debug(f"Executing portal query: {query}")

    if not (items := portal.get_metadata(query, raise_exception=False)):
        return 1

    if not isinstance(items := items.get("@graph"), list):
        _error(f"Query did not return a list as expected: {query}")

    # Example record:
    # {
    #     "software": [
    #         "f349ceba-ac73-4332-b9cb-5cd3b4598d60"
    #     ],
    #     "s3_lifecycle_status": "standard",
    #     "access_status": "Open",
    #     "description": "Extremely-difficult-to-map regions: Regions outside of the easy- or difficult-to-map regions",
    #     "accession": "SMASFP5V44A8",
    #     "uuid": "4a86ba8c-b442-41c2-ab93-516736b98322",
    #     "schema_version": "1",
    #     "reference_genome": "e89937e6-80d3-4605-8dea-4a74c7981a9f",
    #     "md5sum": "7fe0cac3336dc161d315121fde5355fe",
    #     "submission_centers": [
    #         "9626d82e-8110-4213-ac75-0a50adf890ff"
    #     ],
    #     "last_modified": {
    #         "date_modified": "2024-11-01T13:42:31.228341+00:00",
    #         "modified_by": "74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68"
    #     },
    #     "file_format": "4c04f6de-89a7-4477-8dc4-811b50c67401",
    #     "date_created": "2024-08-12T18:12:42.110604+00:00",
    #     "submitted_by": "5519933a-7772-4188-a318-afc84c1302cd",
    #     "data_category": [
    #         "Genome Region"
    #     ],
    #     "content_md5sum": "7fe0cac3336dc161d315121fde5355fe",
    #     "version": "1.0",
    #     "consortia": [
    #         "358aed10-9b9d-4e26-ab84-4bd162da182b"
    #     ],
    #     "submitted_id": "DAC_SUPPLEMENTARY-FILE_GRCH38-EXTREME-REGIONS",
    #     "file_size": 5886352,
    #     "tags": [
    #         "genome_stratification"
    #     ],
    #     "filename": "SMaHT_extreme_regions_GRCh38_v1.0.bed",
    #     "data_type": [
    #         "Sequence Interval"
    #     ],
    #     "dataset": "colo829_snv_indel_challenge_data",
    #     "status": "released"
    # }

    if argv.group:
        grouped_items = _group_by_month(items, date_property_name)
        if argv.dump:
            _print(json.dumps(grouped_items, indent=4))
        else:
            for month in grouped_items:
                items = grouped_items[month]
                _print(f"\n{chars.rarrow} MONTH: {month}")
                _print_file_table(items, date_property_name=date_property_name, argv=argv)
    else:
        if argv.dump:
            _print(json.dumps(grouped_items, indent=4))
        else:
            _print_file_table(items, date_property_name=date_property_name, argv=argv)

    return 0


def _print_file_table(items: List[dict], date_property_name: str, argv: ARGV) -> None:

    table = PrettyTable()
    if argv.verbose:
        table.field_names = ["N", "FILE", "TYPE / SIZE", "STATUS", "CATEGORY", "DATE"]
    else:
        table.field_names = ["N", "FILE", "TYPE / SIZE", "STATUS", "DATE"]
    table.align = "l"
    table.align["N"] = "r"
    if argv.verbose:
        table.hrules = PrettyTableHorizontalStyle.ALL

    nitems = 0

    for item in items:

        nitems += 1
        uuid = item.get("uuid", "")
        type = Portal.get_item_type(item)
        status = item.get("status", "")
        description = item.get("description", "")
        name = item.get("filename", item.get("display_title", ""))
        size = item.get("file_size", "")
        if isinstance(category := item.get("data_category", ""), list):
            if (category := "\n".join(category)) == "Quality Control":
                category = "QC"
        date = get_property(item, date_property_name, get_property(item, "last_modified.date_modified"))
        by = get_property(item, "submitted_by.display_title",
                                get_property(item, "last_modified.modified_by.display_title"))

        type_and_size = f"{type}\n{format_size(size)}"

        if argv.verbose:
            name += f"\n{uuid}"
            name += f"\n{description}"
            date = f"{format_date(date)}\n{format_time(date, notz=True)}"
            if by:
                date += f"\n{by}"
            if type_and_size:
                type_and_size += f"\n{size}"
        else:
            date = f"{format_date(date)}"

        if argv.verbose:
            table.add_row([nitems, name, type_and_size, status, category, date])
        else:
            table.add_row([nitems, name, type_and_size, status, date])

    print(table)


def _parse_from_and_thru_date_args(argv: ARGV) -> Tuple[Optional[str], Optional[str]]:
    from_date = thru_date = None
    if not (argv.from_date and argv.thru_date):
        from_date, thru_date = _get_first_and_last_date_of_month(nmonths=argv.nmonths)
    if argv.from_date:
        if not (from_date := parse_datetime_string(argv.from_date)):
            _error(f"Cannot parse given from date: {argv.from_date}")
    if argv.thru_date:
        if not (thru_date := parse_datetime_string(argv.thru_date)):
            _error(f"Cannot parse given thru date: {argv.thru_date}")
    return format_date(from_date), format_date(thru_date)


def _get_first_and_last_date_of_month(day: Optional[Union[datetime, date]] = None,
                                      nmonths: int = 1) -> Tuple[datetime, datetime]:
    """
    Returns datetime objects in a tuple for the first and last days of the month of the given datetime or date
    argument, or if none is passed then for today. If the nmonths argument is greater than one, then instead of the
    first day of the month of the given day (or today), then returns the first day of the nmonth-th previous month.
    """
    def get_first_date_of_previous_month(day: Optional[Union[datetime]]) -> datetime:
        if not isinstance(day, datetime):
            if isinstance(day, date):
                day = datetime.combine(day, datetime.min.time())
            else:
                day = datetime.today()
        return day.replace(day=1).replace(month=day.month - 1 if day.month > 1 else 12)
    if not isinstance(day, datetime):
        if isinstance(day, date):
            day = datetime.combine(day, datetime.min.time())
        else:
            day = datetime.today()
    first_day = day.replace(day=1)
    next_month = first_day.replace(month=first_day.month % 12 + 1, day=1)
    last_day = next_month - timedelta(days=1)
    if isinstance(nmonths, int) and (nmonths > 1):
        for _ in range(nmonths - 1):
            first_day = get_first_date_of_previous_month(first_day)
    return (first_day, last_day)


def _group_by_month(items: List[dict], date_property_name: str) -> dict:
    grouped_items = {}
    if isinstance(items, list):
        for item in items:
            if ((value := get_property(item, date_property_name)) and (value := parse_datetime_string(value))):
                month = f"{value.year}-{value.month:02}"
                if not grouped_items.get(month):
                    grouped_items[month] = []
                grouped_items[month].append(item)
    return grouped_items


def _print(*args, **kwargs) -> None:
    print(*args, **kwargs)


def _verbose(*args, **kwargs) -> None:
    _print(*args, **kwargs, file=sys.stderr, flush=True)


def _debug(message: str) -> None:
    _print(f"DEBUG: {message}", file=sys.stderr, flush=True)


def _error(message: str, exit: bool = True) -> None:
    print(f"ERROR: {message}", file=sys.stderr, flush=True)
    if exit is not False:
        sys.exit(1)


def _setup_debugging(argv: ARGV) -> None:
    global _verbose, _debug
    if argv.nowarnings: _warning = lambda *args, **kwargs: None  # noqa
    if not argv.verbose: _verbose = lambda *args, **kwargs: None  # noqa
    if not argv.debug: _debug = lambda *args, **kwargs: None  # noqa


if __name__ == "__main__":
    status = main()
    sys.exit(status if isinstance(status, int) else 0)
